#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import hopsworks
from fastmcp import Context
from hopsworks.mcp.models.feature_group import FeatureGroup, FeatureGroups
from hopsworks.mcp.utils.tags import TAGS
from hopsworks_common import client


class FeatureGroupTools:
    """Tools for managing feature groups in Hopsworks MCP."""

    def __init__(self, mcp):
        self.mcp = mcp
        self.mcp.tool(tags=[TAGS.FEATURE_GROUP, TAGS.READ, TAGS.STATEFUL])(
            self.get_feature_groups_in_current_project
        )
        self.mcp.tool(tags=[TAGS.FEATURE_GROUP, TAGS.READ, TAGS.STATELESS])(
            self.get_feature_groups
        )

    def _get_feature_groups(self, feature_store, name: str = None) -> FeatureGroups:
        """
        Helper method to retrieve feature groups from a feature store.

        Args:
            feature_store: The feature store instance.
            name: The name of the feature group to retrieve. If None, retrieves all feature groups.

        Returns:
            str: A string representation of the feature groups.
        """
        try:
            feature_groups = feature_store.get_feature_groups(name=name)
        except TypeError:
            # Handle case where feature_store.get_feature_groups fails due to missing parameters
            # Type error FeatureGroup.__init__() missing 3 required positional arguments: 'name', 'version', and 'featurestore_id'
            # When from_response_json is called without a response
            feature_groups = []

        # If no feature groups are found, return an empty FeatureGroups object
        if not feature_groups:
            return FeatureGroups(feature_groups=[], total=0, offset=0, limit=0)

        if "items" not in feature_groups:
            return FeatureGroup(
                id=feature_groups.id,
                name=feature_groups.name,
                version=feature_groups.version,
                featurestore_id=feature_groups.featurestore_id,
                location=feature_groups.location,
                event_time=feature_groups.event_time,
                online_enabled=feature_groups.online_enabled,
                topic_name=feature_groups.topic_name,
                notification_topic_name=feature_groups.notification_topic_name,
                deprecated=feature_groups.deprecated,
            )

        return FeatureGroups(
            feature_groups=[
                FeatureGroup(
                    id=fg.id,
                    name=fg.name,
                    version=fg.version,
                    featurestore_id=fg.featurestore_id,
                    location=fg.location,
                    event_time=fg.event_time,
                    online_enabled=fg.online_enabled,
                    topic_name=fg.topic_name,
                    notification_topic_name=fg.notification_topic_name,
                    deprecated=fg.deprecated,
                )
                for fg in feature_groups["items"]
            ],
            total=feature_groups["count"],
            offset=0,
            limit=len(feature_groups["items"]),
        )

    async def get_feature_groups_in_current_project(
        self,
        feature_store_project_name: str = None,
        name: str | None = None,
        ctx: Context = None,
    ):
        """Get the feature groups for the current project.

        Args:
            feature_store_project_name: Project name of the feature store. If None, uses the default project feature store.
            name: The name of the feature group to retrieve. If None, retrieves all feature groups

        Returns:
            FeatureGroups: List of feature groups in the current project.
        """
        if ctx:
            await ctx.info("Retrieving feature groups for the current project...")

        # Get the current project and its feature groups
        project = hopsworks.get_current_project()
        feature_store = project.get_feature_store(name=feature_store_project_name)
        return self._get_feature_groups(feature_store, name=name)

    async def get_feature_groups(
        self,
        project_name: str,
        feature_store_project_name: str = None,
        name: str | None = None,
        ctx: Context = None,
    ):
        """Get the feature groups for a specific project.

        Args:
            project_name: The name of the project.
            feature_store_project_name: Project name of the feature store. If None, uses the default project feature store.
            name: The name of the feature group to retrieve. If None, retrieves all feature groups.

        Returns:
            FeatureGroups: List of feature groups in the specified project.
        """
        if ctx:
            await ctx.info(f"Retrieving feature groups for project {project_name}...")

        # Get the specified project and its feature groups
        conn = client.get_connection()
        project = conn.get_project(project_name)
        feature_store = project.get_feature_store(name=feature_store_project_name)
        return self._get_feature_groups(feature_store, name=name)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

from typing import Any, Dict, List

import humps


class Dataset:
    def __init__(
        self,
        id,
        name,
        dataset_type,
        attributes: Dict[str, Any],
        **kwargs,
    ) -> None:
        self._id = id
        self._name = name
        self._description = kwargs.get("description", None)
        self._dataset_type = dataset_type
        self._path = attributes["path"]

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> List[Dataset]:
        json_decamelized = humps.decamelize(json_dict)["items"]
        for dataset in json_decamelized:
            _ = dataset.pop("type", None)
        return [cls(**dataset) for dataset in json_decamelized]

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description

    @property
    def dataset_type(self) -> str:
        return self._dataset_type

    @property
    def path(self) -> str:
        return self._path



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import hopsworks
from fastmcp import Context
from hopsworks.mcp.models.job import Jobs, to_base_model_job
from hopsworks.mcp.utils.tags import TAGS
from hopsworks_common import client


class JobTools:
    """Tools for managing jobs in Hopsworks."""

    def __init__(self, mcp):
        """
        Initialize the JobTools with the MCP server instance.

        Args:
            mcp: The MCP server instance
        """
        self.mcp = mcp
        self.mcp.tool(tags=[TAGS.JOB, TAGS.READ, TAGS.STATEFUL])(
            self.get_jobs_in_current_project
        )
        self.mcp.tool(tags=[TAGS.JOB, TAGS.READ, TAGS.STATELESS])(self.get_jobs)

    async def get_jobs_in_current_project(self, ctx: Context = None):
        """Get the jobs for the current project.

        Returns:
            Jobs: List of jobs in the current project.
        """
        if ctx:
            await ctx.info("Retrieving jobs for the current project...")

        # Get the current project and its jobs
        project = hopsworks.get_current_project()
        jobs = project.get_job_api().get_jobs()

        return Jobs(
            jobs=[to_base_model_job(job) for job in jobs],
            total=jobs["count"] if "count" in jobs else len(jobs),
        )

    async def get_jobs(self, project_name: str, ctx: Context = None):
        """Get the jobs for a specific project.

        Args:
            project_name: The name of the project to retrieve jobs from.

        Returns:
            Jobs: List of jobs in the specified project.
        """
        if ctx:
            await ctx.info(f"Retrieving jobs for project '{project_name}'...")

        # Get the specified project and its jobs
        conn = client.get_connection()
        project = conn.get_project(project_name)
        jobs = project.get_job_api().get_jobs()

        return Jobs(
            jobs=[to_base_model_job(job) for job in jobs],
            total=jobs["count"] if "count" in jobs else len(jobs),
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""Authentication tools for Hopsworks."""

from fastmcp import Context
from hopsworks.mcp.models.project import Project
from hopsworks.mcp.utils.auth import login as hw_login
from hopsworks.mcp.utils.tags import TAGS


class AuthTools:
    """Tools for authenticating with Hopsworks."""

    def __init__(self, mcp):
        """Initialize auth tools.

        Args:
            mcp: The MCP server instance
        """
        self.mcp = mcp

        # Register tools
        self.mcp.tool(tags=[TAGS.AUTH, TAGS.STATEFUL])(self.login)

    async def login(
        self,
        host: str = None,
        port: int = 443,
        project: str = None,
        api_key_value: str = None,
        api_key_file: str = None,
        hostname_verification: bool = False,
        trust_store_path: str = None,
        engine: str = "python",
        ctx: Context = None,
    ) -> Project:
        """Connect to a Hopsworks instance.

        Args:
            host (str): Hopsworks host URL.
            port (int): Hopsworks port (default 443).
            project (str): Project name to access.
            api_key_value (str): API key value for Hopsworks authentication.
            api_key_file (str): Path to a file containing the API key for Hopsworks authentication.
            hostname_verification (bool): Enable hostname verification for Hopsworks authentication.
            trust_store_path (str): Path to the trust store for Hopsworks authentication.
            engine (str): Engine to use (default: python).

        Returns:
            Project: The project details or an error message.
        """
        if ctx:
            await ctx.info(
                f"Connecting to Hopsworks at {host or 'hopsworks.ai'} using {engine} engine..."
            )

        # Perform actual login with the Hopsworks API
        project = hw_login(
            host=host,
            port=port,
            project=project,
            api_key_value=api_key_value,
            api_key_file=api_key_file,
            hostname_verification=hostname_verification,
            trust_store_path=trust_store_path,
            engine=engine,
        )

        return Project(
            name=project.name,
            id=project.id,
            owner=project.owner,
            description=project.description,
            created=project.created,
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import hopsworks
from fastmcp import Context
from hopsworks.mcp.models.dataset import Dataset, Datasets, File, Files
from hopsworks.mcp.utils.tags import TAGS
from hopsworks_common import client
from hopsworks_common.core import dataset, inode


class DatasetTools:
    """Tools for managing datasets in Hopsworks."""

    def __init__(self, mcp):
        """
        Initialize the DatasetTools with the MCP server instance.

        Args:
            mcp: The MCP server instance
        """
        self.mcp = mcp
        self.mcp.tool(tags=[TAGS.DATASET, TAGS.READ, TAGS.STATEFUL])(
            self.get_datasets_in_current_project
        )
        self.mcp.tool(tags=[TAGS.DATASET, TAGS.READ, TAGS.STATELESS])(self.get_datasets)
        self.mcp.tool(tags=[TAGS.DATASET, TAGS.READ, TAGS.STATEFUL])(
            self.list_files_in_current_project
        )
        self.mcp.tool(tags=[TAGS.DATASET, TAGS.READ, TAGS.STATELESS])(self.list_files)
        self.mcp.tool(tags=[TAGS.DATASET, TAGS.WRITE, TAGS.STATEFUL])(
            self.mkdir_in_current_project
        )
        self.mcp.tool(tags=[TAGS.DATASET, TAGS.WRITE, TAGS.STATELESS])(self.mkdir)

    async def get_datasets_in_current_project(
        self,
        offset: int | str = 0,
        limit: int | str = 100,
        ctx: Context = None,
    ) -> Datasets:
        """Get the dataset for the current project.

        Args:
            offset (int | str): The offset for pagination. Defaults to 0.
            limit (int | str): The limit for pagination. Defaults to 100.

        Returns:
            Datasets: The dataset information for the current project or an error message.
        """

        project = hopsworks.get_current_project()
        if ctx:
            await ctx.info(f"Listing files in project {project.name} ...")

        if isinstance(offset, str):
            offset = int(offset)
        if isinstance(limit, str):
            limit = int(limit)

        count, datasets = project.get_dataset_api()._list_dataset_path(
            "", dataset.Dataset, offset=offset, limit=limit
        )

        # Return the dataset details
        return Datasets(
            datasets=[
                Dataset(
                    name=ds.name,
                    id=ds.id,
                    description=ds.description,
                    datasetType=ds.dataset_type,
                )
                for ds in datasets
            ],
            total=count,
            offset=offset,
            limit=limit,
        )

    async def get_datasets(
        self,
        project_name: str,
        offset: int | str = 0,
        limit: int | str = 100,
        ctx: Context = None,
    ) -> Datasets:
        """Get the dataset for a specific project.

        Args:
            project_name (str): The name of the project to get the dataset for.
            offset (int | str): The offset for pagination. Defaults to 0.
            limit (int | str): The limit for pagination. Defaults to 100.

        Returns:
            Datasets: The dataset information for the specified project or an error message.
        """
        if ctx:
            await ctx.info(f"Listing files in project {project_name} ...")

        if isinstance(offset, str):
            offset = int(offset)
        if isinstance(limit, str):
            limit = int(limit)

        conn = client.get_connection()
        project = conn.get_project(project_name)
        count, datasets = project.get_dataset_api()._list_dataset_path(
            "", dataset.Dataset, offset=offset, limit=limit
        )

        # Return the dataset details
        return Datasets(
            datasets=[
                Dataset(
                    name=ds.name,
                    id=ds.id,
                    description=ds.description,
                    datasetType=ds.dataset_type,
                )
                for ds in datasets
            ],
            total=count,
            offset=offset,
            limit=limit,
        )

    async def list_files_in_current_project(
        self,
        path: str,
        offset: int | str = 0,
        limit: int | str = 100,
        ctx: Context = None,
    ) -> Files:
        """List files in a specific path.

        Args:
            path (str): The path to list files in.
            offset (int | str): The offset for pagination. Defaults to 0.
            limit (int | str): The limit for pagination. Defaults to 100.

        Returns:
            Files: List of files in the specified path or an error message.
        """
        if ctx:
            await ctx.info(f"Listing files in path {path} ...")

        if isinstance(offset, str):
            offset = int(offset)
        if isinstance(limit, str):
            limit = int(limit)

        project = hopsworks.get_current_project()
        count, files = project.get_dataset_api()._list_dataset_path(
            path, inode.Inode, offset=offset, limit=limit
        )

        # Return the file details
        return Files(
            files=[
                File(
                    name=f.name,
                    is_directory=f.dir,
                    owner=f.owner,
                    path=f.path,
                    permission=f.permission,
                    last_modified=f.modification_time,
                    under_construction=f.under_construction,
                )
                for f in files
            ],
            total=count,
            offset=offset,
            limit=limit,
        )

    async def list_files(
        self,
        project_name: str,
        path: str,
        offset: int | str = 0,
        limit: int | str = 100,
        ctx: Context = None,
    ) -> Files:
        """List files in a specific path of a project.

        Args:
            project_name (str): The name of the project to list files for.
            path (str): The path to list files in.
            offset (int | str): The offset for pagination. Defaults to 0.
            limit (int | str): The limit for pagination. Defaults to 100.

        Returns:
            Files: List of files in the specified path or an error message.
        """
        if ctx:
            await ctx.info(
                f"Listing files in project {project_name} at path {path} ..."
            )

        if isinstance(offset, str):
            offset = int(offset)
        if isinstance(limit, str):
            limit = int(limit)

        conn = client.get_connection()
        project = conn.get_project(project_name)
        count, files = project.get_dataset_api()._list_dataset_path(
            path, inode.Inode, offset=offset, limit=limit
        )

        # Return the file details
        return Files(
            files=[
                File(
                    name=f.name,
                    is_directory=f.dir,
                    owner=f.owner,
                    path=f.path,
                    permission=f.permission,
                    last_modified=f.modification_time,
                    under_construction=f.under_construction,
                )
                for f in files
            ],
            total=count,
            offset=offset,
            limit=limit,
        )

    async def mkdir_in_current_project(self, path: str, ctx: Context = None) -> str:
        """Create a directory in the current project.

        Args:
            path (str): The path to create the directory in.

        Returns:
            str: Success message or an error message.
        """
        if ctx:
            await ctx.info(f"Creating directory at path {path} ...")

        project = hopsworks.get_current_project()
        project.get_dataset_api().mkdir(path=path)

        return f"Directory created at {path} in the current project."

    async def mkdir(self, project_name: str, path: str, ctx: Context = None) -> str:
        """Create a directory in a specific project.

        Args:
            project_name (str): The name of the project to create the directory in.
            path (str): The path to create the directory in.

        Returns:
            str: Success message or an error message.
        """
        if ctx:
            await ctx.info(
                f"Creating directory at path {path} in project {project_name} ..."
            )

        conn = client.get_connection()
        project = conn.get_project(project_name)
        project.get_dataset_api().mkdir(path=path)

        return f"Directory created at {path} in project {project_name}."



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import copy
import json
import logging
import math
import os
import shutil
import time
from concurrent.futures import ThreadPoolExecutor, wait
from typing import List, Literal, Optional, Type, Union

from hopsworks_common import client, tag, usage, util
from hopsworks_common.client.exceptions import DatasetException, RestAPIError
from hopsworks_common.core import dataset, inode
from tqdm.auto import tqdm


class Chunk:
    def __init__(self, content, number, status):
        self.content = content
        self.number = number
        self.status = status
        self.retries = 0


class DatasetApi:
    def __init__(self):
        self._log = logging.getLogger(__name__)

    DEFAULT_UPLOAD_FLOW_CHUNK_SIZE = 10 * 1024 * 1024
    DEFAULT_UPLOAD_SIMULTANEOUS_UPLOADS = 3
    DEFAULT_UPLOAD_SIMULTANEOUS_CHUNKS = 3
    DEFAULT_UPLOAD_MAX_CHUNK_RETRIES = 1

    DEFAULT_DOWNLOAD_FLOW_CHUNK_SIZE = 1024 * 1024
    FLOW_PERMANENT_ERRORS = [404, 413, 415, 500, 501]

    # alias for backwards-compatibility:
    DEFAULT_FLOW_CHUNK_SIZE = DEFAULT_DOWNLOAD_FLOW_CHUNK_SIZE

    @usage.method_logger
    def download(
        self,
        path: str,
        local_path: Optional[str] = None,
        overwrite: Optional[bool] = False,
        chunk_size: int = DEFAULT_DOWNLOAD_FLOW_CHUNK_SIZE,
    ):
        """Download file from Hopsworks Filesystem to the current working directory.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        downloaded_file_path = dataset_api.download("Resources/my_local_file.txt")

        ```
        # Arguments
            path: path in Hopsworks filesystem to the file
            local_path: path where to download the file in the local filesystem
            overwrite: overwrite local file if exists
            chunk_size: upload chunk size in bytes. Default 1 MB
        # Returns
            `str`: Path to downloaded file
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "dataset",
            "download",
            "with_auth",
            path,
        ]
        query_params = {"type": "DATASET"}

        # Build the path to download the file on the local fs and return to the user, it should be absolute for consistency
        # Download in CWD if local_path not specified
        if local_path is None:
            local_path = os.getcwd()
        # If local_path specified, ensure it is absolute
        elif not os.path.isabs(local_path):
            local_path = os.path.join(os.getcwd(), local_path)

        # If local_path is a directory, download into the directory
        if os.path.isdir(local_path):
            local_path = os.path.join(local_path, os.path.basename(path))

        if overwrite:
            if os.path.isfile(local_path):
                os.remove(local_path)
            elif os.path.isdir(local_path):
                shutil.rmtree(local_path)
        elif os.path.exists(local_path):
            raise IOError(
                "{} already exists, set overwrite=True to overwrite it".format(
                    local_path
                )
            )

        file_size = int(self._get(path)["attributes"]["size"])
        with _client._send_request(
            "GET", path_params, query_params=query_params, stream=True
        ) as response:
            with open(local_path, "wb") as f:
                pbar = None
                try:
                    pbar = tqdm(
                        total=file_size,
                        bar_format="{desc}: {percentage:.3f}%|{bar}| {n_fmt}/{total_fmt} elapsed<{elapsed} remaining<{remaining}",
                        desc="Downloading",
                    )
                except Exception:
                    self._log.exception("Failed to initialize progress bar.")
                    self._log.info("Starting download")

                for chunk in response.iter_content(chunk_size=chunk_size):
                    f.write(chunk)

                    if pbar is not None:
                        pbar.update(len(chunk))

                if pbar is not None:
                    pbar.close()
                else:
                    self._log.info("Download finished")

        return local_path

    @usage.method_logger
    def upload(
        self,
        local_path: str,
        upload_path: str,
        overwrite: bool = False,
        chunk_size: int = DEFAULT_UPLOAD_FLOW_CHUNK_SIZE,
        simultaneous_uploads: int = DEFAULT_UPLOAD_SIMULTANEOUS_UPLOADS,
        simultaneous_chunks: int = DEFAULT_UPLOAD_SIMULTANEOUS_CHUNKS,
        max_chunk_retries: int = DEFAULT_UPLOAD_MAX_CHUNK_RETRIES,
        chunk_retry_interval: int = 1,
    ):
        """Upload a file or directory to the Hopsworks filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        # upload a file to Resources dataset
        uploaded_file_path = dataset_api.upload("my_local_file.txt", "Resources")

        # upload a directory to Resources dataset
        uploaded_file_path = dataset_api.upload("my_dir", "Resources")

        ```
        # Arguments
            local_path: local path to file or directory to upload, can be relative or absolute
            upload_path: path to directory where to upload the file in Hopsworks Filesystem
            overwrite: overwrite file or directory if exists
            chunk_size: upload chunk size in bytes. Default 10 MB
            simultaneous_chunks: number of simultaneous chunks to upload for each file upload. Default 3
            simultaneous_uploads: number of simultaneous files to be uploaded for directories. Default 3
            max_chunk_retries: maximum retry for a chunk. Default is 1
            chunk_retry_interval: chunk retry interval in seconds. Default is 1sec
        # Returns
            `str`: Path to uploaded file or directory
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        # local path could be absolute or relative,
        if not os.path.isabs(local_path) and os.path.exists(
            os.path.join(os.getcwd(), local_path)
        ):
            local_path = os.path.join(os.getcwd(), local_path)

        _, file_name = os.path.split(local_path)

        destination_path = upload_path + "/" + file_name

        if self.exists(destination_path):
            if overwrite:
                if "datasetType" in self._get(destination_path):
                    raise DatasetException(
                        "overwrite=True not supported on a top-level dataset"
                    )
                else:
                    self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        destination_path
                    )
                )

        if os.path.isdir(local_path):
            self.mkdir(destination_path)

        if os.path.isdir(local_path):
            with ThreadPoolExecutor(simultaneous_uploads) as executor:
                # if path is a dir, upload files and folders iteratively
                for root, dirs, files in os.walk(local_path):
                    # os.walk(local_model_path), where local_model_path is expected to be an absolute path
                    # - root is the absolute path of the directory being walked
                    # - dirs is the list of directory names present in the root dir
                    # - files is the list of file names present in the root dir
                    # we need to replace the local path prefix with the hdfs path prefix (i.e., /srv/hops/....../root with /Projects/.../)
                    remote_base_path = root.replace(
                        local_path, destination_path
                    ).replace(os.sep, "/")
                    for d_name in dirs:
                        self.mkdir(remote_base_path + "/" + d_name)

                    # uploading files in the same folder is done concurrently
                    futures = [
                        executor.submit(
                            self._upload_file,
                            f_name,
                            root + os.sep + f_name,
                            remote_base_path,
                            chunk_size,
                            simultaneous_chunks,
                            max_chunk_retries,
                            chunk_retry_interval,
                        )
                        for f_name in files
                    ]

                    # wait for all upload tasks to complete
                    _, _ = wait(futures)
                    try:
                        _ = [future.result() for future in futures]
                    except Exception as e:
                        raise e
        else:
            self._upload_file(
                file_name,
                local_path,
                upload_path,
                chunk_size,
                simultaneous_chunks,
                max_chunk_retries,
                chunk_retry_interval,
            )

        return upload_path + "/" + os.path.basename(local_path)

    def _upload_file(
        self,
        file_name,
        local_path,
        upload_path,
        chunk_size,
        simultaneous_chunks,
        max_chunk_retries,
        chunk_retry_interval,
    ):
        file_size = os.path.getsize(local_path)

        num_chunks = math.ceil(file_size / chunk_size)

        base_params = self._get_flow_base_params(
            file_name, num_chunks, file_size, chunk_size
        )

        chunk_number = 1
        with open(local_path, "rb") as f:
            pbar = None
            try:
                pbar = tqdm(
                    total=file_size,
                    bar_format="{desc}: {percentage:.3f}%|{bar}| {n_fmt}/{total_fmt} elapsed<{elapsed} remaining<{remaining}",
                    desc="Uploading {}".format(local_path),
                )
            except Exception:
                self._log.exception("Failed to initialize progress bar.")
                self._log.info("Starting upload")
            with ThreadPoolExecutor(simultaneous_chunks) as executor:
                all_chunks_processed = False

                while not all_chunks_processed:
                    chunks = []

                    if os.path.getsize(local_path) == 0:
                        chunks.append(Chunk(b"", 1, "pending"))
                        all_chunks_processed = True
                    else:
                        for _ in range(simultaneous_chunks):
                            chunk_data = f.read(chunk_size)
                            if not chunk_data:
                                all_chunks_processed = True
                                break
                            chunks.append(Chunk(chunk_data, chunk_number, "pending"))
                            chunk_number += 1

                    if not chunks:
                        break

                    futures = [
                        executor.submit(
                            self._upload_chunk,
                            base_params,
                            upload_path,
                            file_name,
                            chunk,
                            pbar,
                            max_chunk_retries,
                            chunk_retry_interval,
                        )
                        for chunk in chunks
                    ]

                    try:
                        for future in futures:
                            future.result()
                    except Exception:
                        if pbar:
                            pbar.close()
                        raise

            if pbar is not None:
                pbar.close()
            else:
                self._log.info("Upload finished")

    def _upload_chunk(
        self,
        base_params,
        upload_path,
        file_name,
        chunk: Chunk,
        pbar,
        max_chunk_retries,
        chunk_retry_interval,
    ):
        query_params = copy.copy(base_params)
        query_params["flowCurrentChunkSize"] = len(chunk.content)
        query_params["flowChunkNumber"] = chunk.number

        chunk.status = "uploading"
        while True:
            try:
                self._upload_request(
                    query_params, upload_path, file_name, chunk.content
                )
                break
            except RestAPIError as re:
                chunk.retries += 1
                if (
                    re.response.status_code in DatasetApi.FLOW_PERMANENT_ERRORS
                    or chunk.retries > max_chunk_retries
                ):
                    chunk.status = "failed"
                    raise re
                time.sleep(chunk_retry_interval)
                continue

        chunk.status = "uploaded"

        if pbar is not None:
            pbar.update(query_params["flowCurrentChunkSize"])

    def _get_flow_base_params(self, file_name, num_chunks, size, chunk_size):
        return {
            "templateId": -1,
            "flowChunkSize": chunk_size,
            "flowTotalSize": size,
            "flowIdentifier": str(size) + "_" + file_name,
            "flowFilename": file_name,
            "flowRelativePath": file_name,
            "flowTotalChunks": num_chunks,
        }

    def _upload_request(self, params, path, file_name, chunk):
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", "upload", path]

        # Flow configuration params are sent as form data
        _client._send_request(
            "POST", path_params, data=params, files={"file": (file_name, chunk)}
        )

    def _get(self, path: str):
        """Get dataset metadata.

        # Arguments
            path: path to check
        # Returns
            `dict`: dataset metadata
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", path]
        headers = {"content-type": "application/json"}
        return _client._send_request("GET", path_params, headers=headers)

    def get(self, path: str):
        """**Deprecated**

        Get dataset metadata.

        # Arguments
            path: path to check
        # Returns
            `dict`: dataset metadata
        """
        return self._get(path)

    def exists(self, path: str):
        """Check if a file exists in the Hopsworks Filesystem.

        # Arguments
            path: path to check
        # Returns
            `bool`: True if exists, otherwise False
        """
        try:
            self._get(path)
            return True
        except RestAPIError:
            return False

    def path_exists(self, remote_path: str):
        """**Deprecated**, use `exists` instead.

        Check if a path exists in datasets.

        # Arguments
            remote_path: path to check
        # Returns
            `bool`: True if exists, otherwise False
        """
        return self.exists(remote_path)

    @usage.method_logger
    def remove(self, path: str):
        """Remove a path in the Hopsworks Filesystem.

        # Arguments
            path: path to remove
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", path]
        return _client._send_request("DELETE", path_params)

    def rm(self, remote_path: str):
        """**Deprecated**, use `remove` instead.

        Remove a path in the Hopsworks Filesystem.

        # Arguments
            remote_path: path to remove
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return self.remove(remote_path)

    @usage.method_logger
    def mkdir(self, path: str):
        """Create a directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.mkdir("Resources/my_dir")

        ```
        # Arguments
            path: path to directory
        # Returns
            `str`: Path to created directory
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", path]
        query_params = {
            "action": "create",
            "searchable": "true",
            "generate_readme": "false",
            "type": "DATASET",
        }
        headers = {"content-type": "application/json"}
        return _client._send_request(
            "POST", path_params, headers=headers, query_params=query_params
        )["attributes"]["path"]

    @usage.method_logger
    def copy(self, source_path: str, destination_path: str, overwrite: bool = False):
        """Copy a file or directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.copy("Resources/myfile.txt", "Logs/myfile.txt")

        ```
        # Arguments
            source_path: the source path to copy
            destination_path: the destination path
            overwrite: overwrite destination if exists
        # Raises
            `hopsworks.client.exceptions.DatasetException`: If the destination path already exists and overwrite is not set to True
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        if self.exists(destination_path):
            if overwrite:
                if "datasetType" in self._get(destination_path):
                    raise DatasetException(
                        "overwrite=True not supported on a top-level dataset"
                    )
                else:
                    self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        destination_path
                    )
                )

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", source_path]
        query_params = {
            "action": "copy",
            "destination_path": destination_path,
        }
        _client._send_request("POST", path_params, query_params=query_params)

    @usage.method_logger
    def move(self, source_path: str, destination_path: str, overwrite: bool = False):
        """Move a file or directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.move("Resources/myfile.txt", "Logs/myfile.txt")

        ```
        # Arguments
            source_path: the source path to move
            destination_path: the destination path
            overwrite: overwrite destination if exists
        # Raises
            `hopsworks.client.exceptions.DatasetException`: If the destination path already exists and overwrite is not set to True
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        if self.exists(destination_path):
            if overwrite:
                if "datasetType" in self._get(destination_path):
                    raise DatasetException(
                        "overwrite=True not supported on a top-level dataset"
                    )
                else:
                    self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        destination_path
                    )
                )

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", source_path]
        query_params = {
            "action": "move",
            "destination_path": destination_path,
        }
        _client._send_request("POST", path_params, query_params=query_params)

    @usage.method_logger
    def upload_feature_group(self, feature_group, path, dataframe):
        """Upload a dataframe to a path in Parquet format using a feature group metadata.

        !!! Note
            This method is a legacy method kept for backwards-compatibility; do not use it in new code.
        """
        # Convert the dataframe into PARQUET for upload
        df_parquet = dataframe.to_parquet(index=False)
        parquet_length = len(df_parquet)
        num_chunks = math.ceil(parquet_length / self.DEFAULT_FLOW_CHUNK_SIZE)

        base_params = self._get_flow_base_params(
            feature_group, num_chunks, parquet_length, self.DEFAULT_FLOW_CHUNK_SIZE
        )

        chunk_number = 1
        for i in range(0, parquet_length, self.DEFAULT_FLOW_CHUNK_SIZE):
            query_params = base_params
            query_params["flowCurrentChunkSize"] = len(
                df_parquet[i : i + self.DEFAULT_FLOW_CHUNK_SIZE]
            )
            query_params["flowChunkNumber"] = chunk_number

            self._upload_request(
                query_params,
                path,
                util.feature_group_name(feature_group),
                df_parquet[i : i + self.DEFAULT_FLOW_CHUNK_SIZE],
            )

            chunk_number += 1

    @usage.method_logger
    def list(self, path: str, offset: int = 0, limit: int = 1000) -> List[str]:
        """List the files and directories from a path in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        # list all files in the Resources dataset
        files = dataset_api.list("/Resources")

        # list all datasets in the project
        files = dataset_api.list("/")

        ```
        # Arguments
            path: path in Hopsworks filesystem to the directory
            offset: the number of entities to skip
            limit: max number of the returned entities
        # Returns
            `list[str]`: List of path to files and directories in the provided path
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        # Normalize path so we can check if the path refers to the root or not
        # That is needed as different backend entities are returned depending on if it is a top level dataset or a subdirectory
        normalized_path = os.path.normpath(path)
        if normalized_path == "/":
            normalized_path = ""
            cls = dataset.Dataset
        else:
            cls = inode.Inode

        count, items = self._list_dataset_path(
            normalized_path, cls, offset=offset, limit=limit
        )

        files = []
        for item in items:
            files.append(
                util.convert_to_project_rel_path(item.path, _client._project_name)
            )
        return files

    @usage.method_logger
    def _list_dataset_path(
        self,
        path: str,
        cls: Union[Type[dataset.Dataset], Type[inode.Inode]],
        offset: int = 0,
        limit: int = 1000,
        sort_by: str = "ID:asc",
    ):
        """
        List contents of a directory in the Hopsworks Filesystem.

        # Arguments
            path: path to the directory to list the contents of.
            offset: the number of Inodes to skip.
            limit: max number of the returned Inodes.
        # Returns
            `tuple[int, tuple[int, list[inode.Inode]] | tuple[int, list[Dataset]]]`: count of Dataset or Inodes and objects
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "dataset",
            path,
        ]
        query_params = {
            "action": "listing",
            "offset": offset,
            "limit": limit,
            "sort_by": sort_by,
            "expand": "inodes",
        }

        items = _client._send_request("GET", path_params, query_params)

        return items["count"], cls.from_response_json(items)

    @usage.method_logger
    def read_content(self, path: str, dataset_type: str = "DATASET"):
        """Read the content of a file.

        # Arguments
            path: The path to the file to read.
            dataset_type: The type of dataset, can be `DATASET` or `HIVEDB`; defaults to `DATASET`.
                `HIVEDB` type is used to read files from Apache Hive.

        # Returns
            An object with `content` attribute containing the file content as bytes, or `None` if the file was not found.
        """
        _client = client.get_instance()

        path_params = [
            "project",
            _client._project_id,
            "dataset",
            "download",
            "with_auth",
            path[1:],
        ]

        query_params = {
            "type": dataset_type,
        }

        return _client._send_request("GET", path_params, query_params, stream=True)

    def chmod(self, remote_path: str, permissions: str):
        """Change permissions of a file or a directory in the Hopsworks Filesystem.

        # Arguments
            remote_path: path to change the permissions of.
            permissions: permissions string, for example `"u+x"`.
        # Returns
            `dict`: the updated dataset metadata
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", remote_path]
        headers = {"content-type": "application/json"}
        query_params = {"action": "PERMISSION", "permissions": permissions}
        return _client._send_request(
            "PUT", path_params, headers=headers, query_params=query_params
        )

    # region Archiving

    def _archive(
        self,
        remote_path: str,
        destination_path: Optional[str] = None,
        block: bool = False,
        timeout: Optional[int] = 120,
        action: Union[Literal["unzip"], Literal["zip"]] = "unzip",
    ):
        """Internal (de)compression logic.

        # Arguments
            remote_path: path to file or directory to unzip.
            destination_path: path to upload the zip, defaults to None; is used only if action is zip.
            block: if the operation should be blocking until complete, defaults to False.
            timeout: timeout in seconds for the blocking, defaults to 120; if None, the blocking is unbounded.
            action: zip or unzip, defaults to unzip.

        # Returns
            `bool`: whether the operation completed in the specified timeout; if non-blocking, always returns True.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", remote_path]

        query_params = {"action": action}

        if destination_path is not None:
            query_params["destination_path"] = destination_path
            query_params["destination_type"] = "DATASET"

        headers = {"content-type": "application/json"}

        _client._send_request(
            "POST", path_params, headers=headers, query_params=query_params
        )

        if not block:
            # the call is successful at this point if we don't want to block
            return True

        # Wait for zip file to appear. When it does, check that parent dir zipState is not set to CHOWNING
        count = 0
        while True:
            if action == "zip":
                zip_path = remote_path + ".zip"
                # Get the status of the zipped file
                if destination_path is None:
                    zip_exists = self.path_exists(zip_path)
                else:
                    zip_exists = self.path_exists(
                        destination_path + "/" + os.path.split(zip_path)[1]
                    )
                # Get the zipState of the directory being zipped
                dir_status = self.get(remote_path)
                zip_state = dir_status["zipState"] if "zipState" in dir_status else None
                if zip_exists and zip_state == "NONE":
                    return True
            elif action == "unzip":
                # Get the status of the unzipped dir
                unzipped_dir_exists = self.path_exists(
                    remote_path[: remote_path.index(".")]
                )
                # Get the zipState of the zip being extracted
                dir_status = self.get(remote_path)
                zip_state = dir_status["zipState"] if "zipState" in dir_status else None
                if unzipped_dir_exists and zip_state == "NONE":
                    return True
            time.sleep(1)
            count += 1
            if timeout is not None and count >= timeout:
                self._log.info(
                    f"Timeout of {timeout} seconds exceeded while {action} {remote_path}."
                )
                return False

    def unzip(
        self, remote_path: str, block: bool = False, timeout: Optional[int] = 120
    ):
        """Unzip an archive in the dataset.

        # Arguments
            remote_path: path to file or directory to unzip.
            block: if the operation should be blocking until complete, defaults to False.
            timeout: timeout in seconds for the blocking, defaults to 120; if None, the blocking is unbounded.

        # Returns
            `bool`: whether the operation completed in the specified timeout; if non-blocking, always returns True.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return self._archive(remote_path, block=block, timeout=timeout, action="unzip")

    def zip(
        self,
        remote_path: str,
        destination_path: Optional[str] = None,
        block: bool = False,
        timeout: Optional[int] = 120,
    ):
        """Zip a file or directory in the dataset.

        # Arguments
            remote_path: path to file or directory to unzip.
            destination_path: path to upload the zip, defaults to None.
            block: if the operation should be blocking until complete, defaults to False.
            timeout: timeout in seconds for the blocking, defaults to 120; if None, the blocking is unbounded.

        # Returns
            `bool`: whether the operation completed in the specified timeout; if non-blocking, always returns True.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return self._archive(
            remote_path,
            destination_path=destination_path,
            block=block,
            timeout=timeout,
            action="zip",
        )

    # region Dataset Tags

    def add(self, path: str, name: str, value: str):
        """**Deprecated**

        Attach a name/value tag to a model.

        A tag consists of a name/value pair. Tag names are unique identifiers.
        The value of a tag can be any valid json - primitives, arrays or json objects.

        # Arguments
            path: path to add the tag
            name: name of the tag to be added
            value: value of the tag to be added
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "dataset",
            "tags",
            "schema",
            name,
            path,
        ]
        headers = {"content-type": "application/json"}
        json_value = json.dumps(value)
        _client._send_request("PUT", path_params, headers=headers, data=json_value)

    def delete(self, path: str, name: str):
        """**Deprecated**

        Delete a tag.

        Tag names are unique identifiers.

        # Arguments
            path: path to delete the tags
            name: name of the tag to be removed
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "dataset",
            "tags",
            "schema",
            name,
            path,
        ]
        _client._send_request("DELETE", path_params)

    def get_tags(self, path: str, name: str | None = None):
        """**Deprecated**

        Get the tags.

        Gets all tags if no tag name is specified.

        # Arguments
            path: path to get the tags
            name: tag name

        # Returns
            `dict`: tag names and values
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "dataset",
            "tags",
        ]

        if name is not None:
            path_params.append("schema")
            path_params.append(name)
        else:
            path_params.append("all")

        path_params.append(path)

        return {
            tag._name: json.loads(tag._value)
            for tag in tag.Tag.from_response_json(
                _client._send_request("GET", path_params)
            )
        }



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from typing import Optional

import humps
from hopsworks_common import util


class Creator:
    def __init__(
        self, username=None, firstname=None, lastname=None, email=None, **kwargs
    ):
        self._username = username
        self._firstname = firstname
        self._lastname = lastname
        self._email = email

    @classmethod
    def from_response_json(cls, json_dict: dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def username(self) -> str:
        """Username of the creator."""
        return self._username

    @property
    def firstname(self) -> str:
        """First name of the creator."""
        return self._firstname

    @property
    def lastname(self) -> str:
        """Last name of the creator."""
        return self._lastname

    @property
    def email(self) -> str:
        """Email of the creator."""
        return self._email

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "username": self.username,
            "firstname": self.firstname,
            "lastname": self.lastname,
            "email": self.email,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"Creator(id={self._username}, name={self._firstname} {self._lastname})"


class Tag:
    def __init__(self, key=None, value=None, **kwargs):
        self._key = key
        self._value = value

    @classmethod
    def from_response_json(cls, json_dict: dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def key(self) -> str:
        """Key of the tag."""
        return self._key

    @property
    def value(self) -> str:
        """Value of the tag."""
        return self._value

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "key": self.key,
            "value": self.value,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"Tag(key={self.key}, description={self.value})"


class FeatureHighlights:
    def __init__(self, name=None, description=None, **kwargs):
        self._name = name
        self._description = description

    @classmethod
    def from_response_json(cls, json_dict: dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def name(self) -> str:
        """Name of the feature highlight."""
        return self._name

    @property
    def description(self) -> str:
        """Description of the feature highlight."""
        return self._description

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"FeatureHighlights(name={self.name}, description={self.description})"


class Highlight:
    def __init__(
        self,
        name=None,
        description=None,
        features=None,
        tags=None,
        other_xattrs=None,
        **kwargs,
    ):
        self._name = name
        self._description = description
        self._features = (
            [FeatureHighlights.from_response_json(f) for f in features]
            if features
            else []
        )
        self._tags = [Tag.from_response_json(tag) for tag in tags] if tags else []
        self._other_xattrs = other_xattrs

    @classmethod
    def from_response_json(cls, json_dict: dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def name(self) -> str:
        """Name of the highlight."""
        return self._name

    @property
    def description(self) -> str:
        """Description of the highlight."""
        return self._description

    @property
    def features(self) -> Optional[list]:
        """List of features."""
        return self._features

    @property
    def tags(self) -> Optional[list]:
        """List of tags."""
        return self._tags

    @property
    def other_xattrs(self) -> Optional[dict]:
        """Other xattrs."""
        return self._other_xattrs

    def json(self) -> dict:
        """Convert the object to JSON format."""
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
            "features": [f.to_dict() for f in self.features],
            "tags": [tag.to_dict() for tag in self.tags],
            "other_xattrs": self.other_xattrs,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"Highlight(name={self.name}, description={self.description}, features={self.features}, tags={self.tags}, other_xattrs={self.other_xattrs})"


class FeaturestoreResult:
    def __init__(
        self,
        name=None,
        version=None,
        description=None,
        featurestore_id=None,
        created=None,
        parent_project_id=None,
        parent_project_name=None,
        access_projects=None,
        highlights: Highlight = None,
        creator: Creator = None,
        elastic_id=None,
        **kwargs,
    ):
        self._name = name
        self._version = version
        self._description = description
        self._featurestore_id = featurestore_id
        self._created = created
        self._parent_project_id = parent_project_id
        self._parent_project_name = parent_project_name
        self._access_projects = access_projects
        self._highlights = (
            Highlight.from_response_json(highlights) if highlights else None
        )
        self._creator = Creator.from_response_json(creator) if creator else None
        self._elastic_id = elastic_id

    @classmethod
    def from_response_json(cls, json_dict: dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def name(self) -> str:
        """Name."""
        return self._name

    @property
    def version(self) -> int:
        """Version."""
        return self._version

    @property
    def description(self) -> str:
        """Description."""
        return self._description

    @property
    def featurestore_id(self) -> int:
        """Featurestore ID."""
        return self._featurestore_id

    @property
    def created(self) -> str:
        """Created timestamp."""
        return self._created

    @property
    def parent_project_id(self) -> int:
        """Parent project ID."""
        return self._parent_project_id

    @property
    def parent_project_name(self) -> str:
        """Parent project name."""
        return self._parent_project_name

    @property
    def access_projects(self) -> Optional[dict]:
        """Projects that can access this result."""
        return self._access_projects

    @property
    def highlights(self) -> Optional[Highlight]:
        """Highlights of the result."""
        return self._highlights

    @property
    def creator(self) -> Optional[Creator]:
        """Creator of the result."""
        return self._creator

    @property
    def elastic_id(self) -> str:
        """Elastic ID of the result."""
        return self._elastic_id

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "featurestore_id": self.featurestore_id,
            "created": self.created,
            "parent_project_id": self.parent_project_id,
            "parent_project_name": self.parent_project_name,
            "access_projects": self.access_projects,
            "highlights": self.highlights.to_dict() if self.highlights else None,
            "creator": self.creator.to_dict() if self.creator else None,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"FeaturestoreResult(name={self.name}, version={self.version}, description={self.description}, featurestore_id={self.featurestore_id}, created={self.created}, parent_project_id={self.parent_project_id}, parent_project_name={self.parent_project_name}, access_projects={self.access_projects}, highlights={self.highlights}, creator={self.creator})"


class FeatureResult(FeaturestoreResult):
    def __init__(
        self,
        name=None,
        version=None,
        description=None,
        featurestore_id=None,
        created=None,
        parent_project_id=None,
        parent_project_name=None,
        access_projects=None,
        highlights: Highlight = None,
        creator: Creator = None,
        elastic_id=None,
        featuregroup=None,
        **kwargs,
    ):
        super().__init__(
            name,
            version,
            description,
            featurestore_id,
            created,
            parent_project_id,
            parent_project_name,
            access_projects,
            highlights,
            creator,
            elastic_id,
            **kwargs,
        )
        from hsfs.core import feature_group_api

        self._featuregroup = featuregroup
        self._feature_group_api = feature_group_api.FeatureGroupApi()

    def get_feature_group(self):
        """
        Get the feature group instance associated with this feature.

        # Returns:
            FeatureGroup: The feature group instance.
        # Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return self._feature_group_api.get(
            self._featurestore_id, self._featuregroup, self._version
        )

    def get_feature(self):
        """
        Get the feature instance associated with this feature result.

        # Returns:
            Feature: The feature instance.
        # Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        fg = self._feature_group_api.get(
            self._featurestore_id, self._featuregroup, self._version
        )
        return fg.get_feature(self.name)

    @property
    def featuregroup(self) -> str:
        """Feature group name."""
        return self._featuregroup

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "featurestore_id": self.featurestore_id,
            "created": self.created,
            "parent_project_id": self.parent_project_id,
            "parent_project_name": self.parent_project_name,
            "access_projects": self.access_projects,
            "highlights": self.highlights.to_dict() if self.highlights else None,
            "creator": self.creator.to_dict() if self.creator else None,
            "featuregroup": self.featuregroup,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"FeatureResult(name={self.name}, version={self.version}, description={self.description}, featurestore_id={self.featurestore_id}, created={self.created}, parent_project_id={self.parent_project_id}, parent_project_name={self.parent_project_name}, access_projects={self.access_projects}, highlights={self.highlights}, creator={self.creator}, featuregroup={self.featuregroup})"


class FeatureGroupResult(FeaturestoreResult):
    def __init__(
        self,
        name=None,
        version=None,
        description=None,
        featurestore_id=None,
        created=None,
        parent_project_id=None,
        parent_project_name=None,
        access_projects=None,
        highlights: Highlight = None,
        creator: Creator = None,
        elastic_id=None,
        **kwargs,
    ):
        super().__init__(
            name,
            version,
            description,
            featurestore_id,
            created,
            parent_project_id,
            parent_project_name,
            access_projects,
            highlights,
            creator,
            elastic_id,
            **kwargs,
        )

    def get_feature_group(self):
        """
        Get the feature group instance associated with this feature group result.

        # Returns:
            FeatureGroup: The feature group instance.
        # Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        from hsfs.core import feature_group_api

        return feature_group_api.FeatureGroupApi().get(
            self._featurestore_id, self._name, self._version
        )

    def __repr__(self) -> str:
        return f"FeatureGroupResult(name={self.name}, version={self.version}, description={self.description}, featurestore_id={self.featurestore_id}, created={self.created}, parent_project_id={self.parent_project_id}, parent_project_name={self.parent_project_name}, access_projects={self.access_projects}, highlights={self.highlights}, creator={self.creator})"


class FeatureViewResult(FeaturestoreResult):
    def __init__(
        self,
        name=None,
        version=None,
        description=None,
        featurestore_id=None,
        created=None,
        parent_project_id=None,
        parent_project_name=None,
        access_projects=None,
        highlights: Highlight = None,
        creator: Creator = None,
        elastic_id=None,
        **kwargs,
    ):
        super().__init__(
            name,
            version,
            description,
            featurestore_id,
            created,
            parent_project_id,
            parent_project_name,
            access_projects,
            highlights,
            creator,
            elastic_id,
            **kwargs,
        )

    def get_feature_view(self):
        """
        Get the feature view instance associated with this feature view result.

        # Returns:
            FeatureView: The feature view instance.
        # Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        from hsfs.core import feature_view_api

        return feature_view_api.FeatureViewApi(
            self._featurestore_id
        ).get_by_name_version(self._name, self._version)

    def __repr__(self) -> str:
        return f"FeatureViewResult(name={self.name}, version={self.version}, description={self.description}, featurestore_id={self.featurestore_id}, created={self.created}, parent_project_id={self.parent_project_id}, parent_project_name={self.parent_project_name}, access_projects={self.access_projects}, highlights={self.highlights}, creator={self.creator})"


class TrainingDatasetResult(FeaturestoreResult):
    def __init__(
        self,
        name=None,
        version=None,
        description=None,
        featurestore_id=None,
        created=None,
        parent_project_id=None,
        parent_project_name=None,
        access_projects=None,
        highlights: Highlight = None,
        creator: Creator = None,
        elastic_id=None,
        **kwargs,
    ):
        super().__init__(
            name,
            version,
            description,
            featurestore_id,
            created,
            parent_project_id,
            parent_project_name,
            access_projects,
            highlights,
            creator,
            elastic_id,
            **kwargs,
        )

    def get_training_dataset(self):
        """
        Get the training dataset instance associated with this training dataset result.

        # Returns:
            TrainingDataset: The training dataset instance.
        # Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        from hsfs.core import training_dataset_api

        return training_dataset_api.TrainingDatasetApi(self._featurestore_id).get(
            self._name, self._version
        )

    def __repr__(self) -> str:
        return f"TrainingDatasetResult(name={self.name}, version={self.version}, description={self.description}, featurestore_id={self.featurestore_id}, created={self.created}, parent_project_id={self.parent_project_id}, parent_project_name={self.parent_project_name}, access_projects={self.access_projects}, highlights={self.highlights}, creator={self.creator})"


class FeaturestoreSearchResultBase:
    def __init__(
        self,
        featuregroups=None,
        feature_views=None,
        trainingdatasets=None,
        features=None,
        featuregroups_from=None,
        featuregroups_total=None,
        feature_views_from=None,
        feature_views_total=None,
        trainingdatasets_from=None,
        trainingdatasets_total=None,
        features_from=None,
        features_total=None,
        **kwargs,
    ):
        self._featuregroups = featuregroups
        self._feature_views = feature_views
        self._trainingdatasets = trainingdatasets
        self._features = features
        self._featuregroups_from = featuregroups_from
        self._featuregroups_total = featuregroups_total
        self._feature_views_from = feature_views_from
        self._feature_views_total = feature_views_total
        self._trainingdatasets_from = trainingdatasets_from
        self._trainingdatasets_total = trainingdatasets_total
        self._features_from = features_from
        self._features_total = features_total

    @classmethod
    def from_response_json(cls, json_dict: dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def featuregroups(self) -> Optional[list]:
        """List of FeatureGroupResult."""
        return self._featuregroups

    @property
    def feature_views(self) -> Optional[list]:
        """List of FeatureViewResult."""
        return self._feature_views

    @property
    def trainingdatasets(self) -> Optional[list]:
        """List of TrainingDatasetResult."""
        return self._trainingdatasets

    @property
    def features(self) -> Optional[list]:
        """List of FeatureResult."""
        return self._features

    @property
    def featuregroups_from(self) -> Optional[list]:
        """Results from offset."""
        return self._featuregroups_from

    @property
    def featuregroups_total(self) -> Optional[list]:
        """Total feature groups found."""
        return self._featuregroups_total

    @property
    def feature_views_from(self) -> Optional[list]:
        """Results from offset."""
        return self._feature_views_from

    @property
    def feature_views_total(self) -> Optional[list]:
        """Total feature views found."""
        return self._feature_views_total

    @property
    def trainingdatasets_from(self) -> Optional[list]:
        """Results from offset."""
        return self._trainingdatasets_from

    @property
    def trainingdatasets_total(self) -> Optional[list]:
        """Total training datasets found."""
        return self._trainingdatasets_total

    @property
    def features_from(self) -> Optional[list]:
        """Features from offset."""
        return self._features_from

    @property
    def features_total(self) -> Optional[list]:
        """Total features found."""
        return self._features_total

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "featuregroups": [fg.to_dict() for fg in self.featuregroups],
            "feature_views": [fv.to_dict() for fv in self.feature_views],
            "trainingdatasets": [td.to_dict() for td in self.trainingdatasets],
            "features": [f.to_dict() for f in self.features],
            "featuregroups_from": self.featuregroups_from,
            "featuregroups_total": self.featuregroups_total,
            "feature_views_from": self.feature_views_from,
            "feature_views_total": self.feature_views_total,
            "trainingdatasets_from": self.trainingdatasets_from,
            "trainingdatasets_total": self.trainingdatasets_total,
            "features_from": self.features_from,
            "features_total": self.features_total,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"FeaturestoreSearchResult(featuregroups={self.featuregroups}, feature_views={self.feature_views}, trainingdatasets={self.trainingdatasets}, features={self.features}, featuregroups_from={self.featuregroups_from}, featuregroups_total={self.featuregroups_total}, feature_views_from={self.feature_views_from}, feature_views_total={self.feature_views_total}, trainingdatasets_from={self.trainingdatasets_from}, trainingdatasets_total={self.trainingdatasets_total}, features_from={self.features_from}, features_total={self.features_total})"


class FeaturestoreSearchResult(FeaturestoreSearchResultBase):
    def __init__(
        self,
        featuregroups=None,
        feature_views=None,
        trainingdatasets=None,
        features=None,
        featuregroups_from=None,
        featuregroups_total=None,
        feature_views_from=None,
        feature_views_total=None,
        trainingdatasets_from=None,
        trainingdatasets_total=None,
        features_from=None,
        features_total=None,
        **kwargs,
    ):
        _featuregroups = (
            [FeatureGroupResult.from_response_json(fg) for fg in featuregroups]
            if featuregroups
            else []
        )
        _feature_views = (
            [FeatureViewResult.from_response_json(fv) for fv in feature_views]
            if feature_views
            else []
        )
        _trainingdatasets = (
            [TrainingDatasetResult.from_response_json(td) for td in trainingdatasets]
            if trainingdatasets
            else []
        )
        _features = (
            [FeatureResult.from_response_json(f) for f in features] if features else []
        )
        super().__init__(
            _featuregroups,
            _feature_views,
            _trainingdatasets,
            _features,
            featuregroups_from,
            featuregroups_total,
            feature_views_from,
            feature_views_total,
            trainingdatasets_from,
            trainingdatasets_total,
            features_from,
            features_total,
        )


class FeaturestoreSearchResultByTag(FeaturestoreSearchResultBase):
    def __init__(
        self,
        featuregroups=None,
        feature_views=None,
        trainingdatasets=None,
        features=None,
        featuregroups_from=None,
        featuregroups_total=None,
        feature_views_from=None,
        feature_views_total=None,
        trainingdatasets_from=None,
        trainingdatasets_total=None,
        features_from=None,
        features_total=None,
        **kwargs,
    ):
        _featuregroups = (
            [
                FeatureGroupResult.from_response_json(fg)
                for fg in featuregroups
                if "highlights" in fg
                and "tags" in fg["highlights"]
                and len(fg["highlights"]["tags"]) > 0
            ]
            if featuregroups
            else []
        )
        _feature_views = (
            [
                FeatureViewResult.from_response_json(fv)
                for fv in feature_views
                if "highlights" in fv
                and "tags" in fv["highlights"]
                and len(fv["highlights"]["tags"]) > 0
            ]
            if feature_views
            else []
        )
        _trainingdatasets = (
            [
                TrainingDatasetResult.from_response_json(td)
                for td in trainingdatasets
                if "highlights" in td
                and "tags" in td["highlights"]
                and len(td["highlights"]["tags"]) > 0
            ]
            if trainingdatasets
            else []
        )
        _features = (
            [
                FeatureResult.from_response_json(f)
                for f in features
                if "highlights" in f
                and "tags" in f["highlights"]
                and len(f["highlights"]["tags"]) > 0
            ]
            if features
            else []
        )
        super().__init__(
            _featuregroups,
            _feature_views,
            _trainingdatasets,
            _features,
            featuregroups_from,
            featuregroups_total,
            feature_views_from,
            feature_views_total,
            trainingdatasets_from,
            trainingdatasets_total,
            features_from,
            features_total,
        )


class FeaturestoreSearchResultByTagName(FeaturestoreSearchResultBase):
    def __init__(
        self,
        featuregroups=None,
        feature_views=None,
        trainingdatasets=None,
        features=None,
        featuregroups_from=None,
        featuregroups_total=None,
        feature_views_from=None,
        feature_views_total=None,
        trainingdatasets_from=None,
        trainingdatasets_total=None,
        features_from=None,
        features_total=None,
        **kwargs,
    ):
        _featuregroups = (
            [
                FeatureGroupResult.from_response_json(fg)
                for fg in featuregroups
                if "highlights" in fg
                and "tags" in fg["highlights"]
                and any(
                    "key" in tag and tag["key"] and "<em>" in tag["key"]
                    for tag in fg["highlights"]["tags"]
                )
            ]
            if featuregroups
            else []
        )
        _feature_views = (
            [
                FeatureViewResult.from_response_json(fv)
                for fv in feature_views
                if "highlights" in fv
                and "tags" in fv["highlights"]
                and any(
                    "key" in tag and tag["key"] and "<em>" in tag["key"]
                    for tag in fv["highlights"]["tags"]
                )
            ]
            if feature_views
            else []
        )
        _trainingdatasets = (
            [
                TrainingDatasetResult.from_response_json(td)
                for td in trainingdatasets
                if "highlights" in td
                and "tags" in td["highlights"]
                and any(
                    "key" in tag and tag["key"] and "<em>" in tag["key"]
                    for tag in td["highlights"]["tags"]
                )
            ]
            if trainingdatasets
            else []
        )
        _features = (
            [
                FeatureResult.from_response_json(f)
                for f in features
                if "highlights" in f
                and "tags" in f["highlights"]
                and any(
                    "key" in tag and tag["key"] and "<em>" in tag["key"]
                    for tag in f["highlights"]["tags"]
                )
            ]
            if features
            else []
        )
        super().__init__(
            _featuregroups,
            _feature_views,
            _trainingdatasets,
            _features,
            featuregroups_from,
            featuregroups_total,
            feature_views_from,
            feature_views_total,
            trainingdatasets_from,
            trainingdatasets_total,
            features_from,
            features_total,
        )


class FeaturestoreSearchResultByTagKey(FeaturestoreSearchResultBase):
    def __init__(
        self,
        featuregroups=None,
        feature_views=None,
        trainingdatasets=None,
        features=None,
        featuregroups_from=None,
        featuregroups_total=None,
        feature_views_from=None,
        feature_views_total=None,
        trainingdatasets_from=None,
        trainingdatasets_total=None,
        features_from=None,
        features_total=None,
        **kwargs,
    ):
        _featuregroups = (
            [
                FeatureGroupResult.from_response_json(fg)
                for fg in featuregroups
                if "highlights" in fg
                and "tags" in fg["highlights"]
                and any(
                    [
                        "<em>" in val.split(":")[0]
                        for tag in fg["highlights"]["tags"]
                        if "value" in tag and isinstance(tag["value"], str)
                        for val in tag["value"].split(",")
                        if ":" in val
                    ]
                )
            ]
            if featuregroups
            else []
        )
        _feature_views = (
            [
                FeatureViewResult.from_response_json(fv)
                for fv in feature_views
                if "highlights" in fv
                and "tags" in fv["highlights"]
                and any(
                    "value" in tag and tag["value"] for tag in fv["highlights"]["tags"]
                )
                and any(
                    [
                        "<em>" in val.split(":")[0]
                        for tag in fv["highlights"]["tags"]
                        if "value" in tag and isinstance(tag["value"], str)
                        for val in tag["value"].split(",")
                        if ":" in val
                    ]
                )
            ]
            if feature_views
            else []
        )
        _trainingdatasets = (
            [
                TrainingDatasetResult.from_response_json(td)
                for td in trainingdatasets
                if "highlights" in td
                and "tags" in td["highlights"]
                and any(
                    "value" in tag and tag["value"] for tag in td["highlights"]["tags"]
                )
                and any(
                    [
                        "<em>" in val.split(":")[0]
                        for tag in td["highlights"]["tags"]
                        if "value" in tag and isinstance(tag["value"], str)
                        for val in tag["value"].split(",")
                        if ":" in val
                    ]
                )
            ]
            if trainingdatasets
            else []
        )
        _features = (
            [
                FeatureResult.from_response_json(f)
                for f in features
                if "highlights" in f
                and "tags" in f["highlights"]
                and any(
                    "value" in tag and tag["value"] for tag in f["highlights"]["tags"]
                )
                and any(
                    [
                        "<em>" in val.split(":")[0]
                        for tag in f["highlights"]["tags"]
                        if "value" in tag and isinstance(tag["value"], str)
                        for val in tag["value"].split(",")
                        if ":" in val
                    ]
                )
            ]
            if features
            else []
        )
        super().__init__(
            _featuregroups,
            _feature_views,
            _trainingdatasets,
            _features,
            featuregroups_from,
            featuregroups_total,
            feature_views_from,
            feature_views_total,
            trainingdatasets_from,
            trainingdatasets_total,
            features_from,
            features_total,
        )


class FeaturestoreSearchResultByTagValue(FeaturestoreSearchResultBase):
    def __init__(
        self,
        featuregroups=None,
        feature_views=None,
        trainingdatasets=None,
        features=None,
        featuregroups_from=None,
        featuregroups_total=None,
        feature_views_from=None,
        feature_views_total=None,
        trainingdatasets_from=None,
        trainingdatasets_total=None,
        features_from=None,
        features_total=None,
        **kwargs,
    ):
        _featuregroups = (
            [
                FeatureGroupResult.from_response_json(fg)
                for fg in featuregroups
                if "highlights" in fg
                and "tags" in fg["highlights"]
                and any(
                    [
                        "<em>" in val.split(":")[1]
                        for tag in fg["highlights"]["tags"]
                        if "value" in tag and isinstance(tag["value"], str)
                        for val in tag["value"].split(",")
                        if ":" in val
                    ]
                )
            ]
            if featuregroups
            else []
        )
        _feature_views = (
            [
                FeatureViewResult.from_response_json(fv)
                for fv in feature_views
                if "highlights" in fv
                and "tags" in fv["highlights"]
                and any(
                    [
                        "<em>" in val.split(":")[1]
                        for tag in fv["highlights"]["tags"]
                        if "value" in tag and isinstance(tag["value"], str)
                        for val in tag["value"].split(",")
                        if ":" in val
                    ]
                )
            ]
            if feature_views
            else []
        )
        _trainingdatasets = (
            [
                TrainingDatasetResult.from_response_json(td)
                for td in trainingdatasets
                if "highlights" in td
                and "tags" in td["highlights"]
                and any(
                    [
                        "<em>" in val.split(":")[1]
                        for tag in td["highlights"]["tags"]
                        if "value" in tag and isinstance(tag["value"], str)
                        for val in tag["value"].split(",")
                        if ":" in val
                    ]
                )
            ]
            if trainingdatasets
            else []
        )
        _features = (
            [
                FeatureResult.from_response_json(f)
                for f in features
                if "highlights" in f
                and "tags" in f["highlights"]
                and any(
                    [
                        "<em>" in val.split(":")[1]
                        for tag in f["highlights"]["tags"]
                        if "value" in tag and isinstance(tag["value"], str)
                        for val in tag["value"].split(",")
                        if ":" in val
                    ]
                )
            ]
            if features
            else []
        )
        super().__init__(
            _featuregroups,
            _feature_views,
            _trainingdatasets,
            _features,
            featuregroups_from,
            featuregroups_total,
            feature_views_from,
            feature_views_total,
            trainingdatasets_from,
            trainingdatasets_total,
            features_from,
            features_total,
        )


class FeaturestoreSearchResultByKeyWord(FeaturestoreSearchResultBase):
    def __init__(
        self,
        featuregroups=None,
        feature_views=None,
        trainingdatasets=None,
        features=None,
        featuregroups_from=None,
        featuregroups_total=None,
        feature_views_from=None,
        feature_views_total=None,
        trainingdatasets_from=None,
        trainingdatasets_total=None,
        features_from=None,
        features_total=None,
        **kwargs,
    ):
        _featuregroups = (
            [
                FeatureGroupResult.from_response_json(fg)
                for fg in featuregroups
                if "highlights" in fg
                and "other_xattrs" in fg["highlights"]
                and "xattr.keywords" in fg["highlights"]["other_xattrs"]
            ]
            if featuregroups
            else []
        )
        _feature_views = (
            [
                FeatureViewResult.from_response_json(fv)
                for fv in feature_views
                if "highlights" in fv
                and "other_xattrs" in fv["highlights"]
                and "xattr.keywords" in fv["highlights"]["other_xattrs"]
            ]
            if feature_views
            else []
        )
        _trainingdatasets = (
            [
                TrainingDatasetResult.from_response_json(td)
                for td in trainingdatasets
                if "highlights" in td
                and "other_xattrs" in td["highlights"]
                and "xattr.keywords" in td["highlights"]["other_xattrs"]
            ]
            if trainingdatasets
            else []
        )
        _features = (
            [
                FeatureResult.from_response_json(f)
                for f in features
                if "highlights" in f
                and "other_xattrs" in f["highlights"]
                and "xattr.keywords" in f["highlights"]["other_xattrs"]
            ]
            if features
            else []
        )
        super().__init__(
            _featuregroups,
            _feature_views,
            _trainingdatasets,
            _features,
            featuregroups_from,
            featuregroups_total,
            feature_views_from,
            feature_views_total,
            trainingdatasets_from,
            trainingdatasets_total,
            features_from,
            features_total,
        )


class FeatureGroupSearchResult:
    def __init__(self, result: FeaturestoreSearchResultBase):
        self._featuregroups = result.featuregroups
        self._featuregroups_from = result.featuregroups_from
        self._featuregroups_total = result.featuregroups_total

    @property
    def featuregroups(self) -> Optional[list]:
        """List of FeatureGroupResult."""
        return self._featuregroups

    @property
    def featuregroups_from(self) -> Optional[list]:
        """Result from offset."""
        return self._featuregroups_from

    @property
    def featuregroups_total(self) -> Optional[list]:
        """Total feature groups found."""
        return self._featuregroups_total

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "featuregroups": [fg.to_dict() for fg in self.featuregroups],
            "featuregroups_from": self.featuregroups_from,
            "featuregroups_total": self.featuregroups_total,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"FeatureGroupSearchResult(featuregroups={self.featuregroups}, featuregroups_from={self.featuregroups_from}, featuregroups_total={self.featuregroups_total})"


class FeatureViewSearchResult:
    def __init__(self, result: FeaturestoreSearchResultBase):
        self._feature_views = result.feature_views
        self._feature_views_from = result.feature_views_from
        self._feature_views_total = result.feature_views_total

    @property
    def feature_views(self) -> Optional[list]:
        """List of FeatureViewResult."""
        return self._feature_views

    @property
    def feature_views_from(self) -> Optional[list]:
        """Result from offset."""
        return self._feature_views_from

    @property
    def feature_views_total(self) -> Optional[list]:
        """Total feature views found."""
        return self._feature_views_total

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "feature_views": [fg.to_dict() for fg in self.feature_views],
            "feature_views_from": self.feature_views_from,
            "feature_views_total": self.feature_views_total,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"FeatureViewSearchResult(feature_views={self.feature_views}, feature_views_from={self.feature_views_from}, feature_views_total={self.feature_views_total})"


class FeatureSearchResult:
    def __init__(self, result: FeaturestoreSearchResultBase):
        self._features = result.features
        self._features_from = result.features_from
        self._features_total = result.features_total

    @property
    def features(self) -> Optional[list]:
        """List of FeatureResult."""
        return self._features

    @property
    def features_from(self) -> Optional[list]:
        """Result from offset."""
        return self._features_from

    @property
    def features_total(self) -> Optional[list]:
        """Total features found."""
        return self._features_total

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "features": [fg.to_dict() for fg in self.features],
            "features_from": self.features_from,
            "features_total": self.features_total,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"FeatureSearchResult(features={self.features}, features_from={self.features_from}, features_total={self.features_total})"


class TrainingDatasetSearchResult:
    def __init__(self, result: FeaturestoreSearchResultBase):
        self._trainingdatasets = result.trainingdatasets
        self._trainingdatasets_from = result.trainingdatasets_from
        self._trainingdatasets_total = result.trainingdatasets_total

    @property
    def trainingdatasets(self) -> Optional[list]:
        """List of TrainingDatasetResult."""
        return self._trainingdatasets

    @property
    def trainingdatasets_from(self) -> Optional[list]:
        """Result from offset."""
        return self._trainingdatasets_from

    @property
    def trainingdatasets_total(self) -> Optional[list]:
        """Total training datasets found."""
        return self._trainingdatasets_total

    def json(self) -> dict:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        return {
            "trainingdatasets": [td.to_dict() for td in self.trainingdatasets],
            "trainingdatasets_from": self.trainingdatasets_from,
            "trainingdatasets_total": self.trainingdatasets_total,
        }

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"TrainingdatasetsSearchResult(trainingdatasets={self.trainingdatasets}, trainingdatasets_from={self.trainingdatasets_from}, trainingdatasets_total={self.trainingdatasets_total})"



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hsml.constants import PREDICTOR
from hsml.predictor import Predictor


class Endpoint(Predictor):
    """Configuration for an endpoint running a python script."""

    def __init__(self, name: str, script_file: str, **kwargs):
        kwargs["model_server"] = PREDICTOR.MODEL_SERVER_PYTHON

        if name is None:
            raise ValueError("A deployment name must be provided")
        if script_file is None:
            raise ValueError(
                "Python scripts are required in deployments without models"
            )

        super().__init__(name=name, script_file=script_file, **kwargs)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import logging
from typing import Literal, get_args

from hopsworks_common import client, search_results


DOC_TYPE_ARG = Literal[
    "FEATUREGROUP", "FEATUREVIEW", "TRAININGDATASET", "FEATURE", "ALL"
]
FILTER_BY_ARG = Literal["tag", "tag_name", "tag_key", "tag_value", "keyword"]


class SearchApi:
    def __init__(self):
        self._log = logging.getLogger(__name__)

    def featurestore_search(
        self,
        search_term: str,
        filter_by: FILTER_BY_ARG = None,
        offset: int = 0,
        limit: int = 100,
    ):
        """
        Search for feature groups, feature views, training datasets and features.
        Returns matching results from all feature stores in the project including shared feature stores.

        ```python

        import hopsworks

        project = hopsworks.login()

        search_api = project.get_search_api()

        result = search_api.featurestore_search("search-term")

        # get feature group instance
        featuregroup = result.featuregroups[0].get_feature_group()

        ```
        # Arguments
            search_term: the term to search for.
            filter_by: filter results by a specific field (default is None).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
        # Returns
            `FeaturestoreSearchResult`: The results.
        # Raises
            `ValueError`: If the search term is not provided or if the filter_by is not one of the allowed values.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return self._search(
            search_term, "ALL", filter_by=filter_by, offset=offset, limit=limit
        )

    # FEATUREGROUP
    def featuregroup_search(
        self,
        search_term: str,
        filter_by: FILTER_BY_ARG = None,
        offset: int = 0,
        limit: int = 100,
    ):
        """
        Search for feature group.
        Returns matching feature groups from all feature stores in the project including shared feature stores.

        ```python

        import hopsworks

        project = hopsworks.login()

        search_api = project.get_search_api()

        result = search_api.featuregroup_search("search_term")

        # get feature group instance
        featuregroup = result.featuregroups[0].get_feature_group()

        ```
        # Arguments
            search_term: the term to search for.
            filter_by: filter results by a specific field (default is None).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
        # Returns
            `FeatureGroupSearchResult`: The results.
        # Raises
            `ValueError`: If the search term is not provided or if the filter_by is not one of the allowed values.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return search_results.FeatureGroupSearchResult(
            self._search(
                search_term,
                "FEATUREGROUP",
                filter_by=filter_by,
                offset=offset,
                limit=limit,
            )
        )

    # FEATUREVIEW
    def featureview_search(
        self,
        search_term: str,
        filter_by: FILTER_BY_ARG = None,
        offset: int = 0,
        limit: int = 100,
    ):
        """
        Search for feature views.
        Returns matching feature views from all feature stores in the project including shared feature stores.

        ```python

        import hopsworks

        project = hopsworks.login()

        search_api = project.get_search_api()

        result = search_api.featureview_search("search_term")

        # get feature view instance
        featureview = result.feature_views[0].get_feature_view()

        ```
        # Arguments
            search_term: the term to search for.
            filter_by: filter results by a specific field (default is None).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
        # Returns
            `FeatureViewSearchResult`: The results.
        # Raises
            `ValueError`: If the search term is not provided or if the filter_by is not one of the allowed values.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return search_results.FeatureViewSearchResult(
            self._search(
                search_term,
                "FEATUREVIEW",
                filter_by=filter_by,
                offset=offset,
                limit=limit,
            )
        )

    # TRAININGDATASET
    def trainingdataset_search(
        self,
        search_term: str,
        filter_by: FILTER_BY_ARG = None,
        offset: int = 0,
        limit: int = 100,
    ):
        """
        Search for training datasets.
        Returns matching training datasets from all feature stores in the project including shared feature stores.

        ```python

        import hopsworks

        project = hopsworks.login()

        search_api = project.get_search_api()

        result = search_api.trainingdataset_search("search_term")

        # get training datasets instance
        trainingdataset = result.trainingdatasets[0].get_training_dataset()

        ```
        # Arguments
            search_term: the term to search for.
            filter_by: filter results by a specific field (default is None).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
        # Returns
            `TrainingDatasetSearchResult`: The results.
        # Raises
            `ValueError`: If the search term is not provided or if the filter_by is not one of the allowed values.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return search_results.TrainingDatasetSearchResult(
            self._search(
                search_term,
                "TRAININGDATASET",
                filter_by=filter_by,
                offset=offset,
                limit=limit,
            )
        )

    # FEATURE
    def feature_search(
        self,
        search_term: str,
        offset: int = 0,
        limit: int = 100,
    ):
        """
        Search for features.
        Returns matching features from all feature stores in the project including shared feature stores.

        ```python

        import hopsworks

        project = hopsworks.login()

        search_api = project.get_search_api()

        result = search_api.feature_search("search_term")

        # get feature group instance
        feature_group = result.features[0].get_feature_group()

        # get feature instance
        feature = result.features[0].get_feature()

        ```
        # Arguments
            search_term: the term to search for.
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
        # Returns
            `FeatureSearchResult`: The results.
        # Raises
            `ValueError`: If the search term is not provided.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        result = self._search(
            search_term, "FEATURE", filter_by=None, offset=offset, limit=limit
        )

        return search_results.FeatureSearchResult(result)

    def _search(
        self,
        search_term: str,
        doc_type: DOC_TYPE_ARG,
        filter_by: FILTER_BY_ARG,
        offset: int,
        limit: int,
    ):
        if not search_term:
            raise ValueError("Search term not provided.")
        if doc_type not in get_args(DOC_TYPE_ARG):
            raise ValueError(
                f"doc_type must be one of the following {get_args(DOC_TYPE_ARG)}."
            )
        if filter_by is not None and filter_by not in get_args(FILTER_BY_ARG):
            raise ValueError(
                f"filter_by must be one of the following {get_args(FILTER_BY_ARG)}."
            )

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "elastic",
            "featurestore",
            search_term,
        ]
        headers = {"content-type": "application/json"}
        query_params = {"docType": doc_type, "from": offset, "size": limit}

        result = _client._send_request(
            "GET", path_params, query_params=query_params, headers=headers
        )

        if filter_by == "tag":
            return search_results.FeaturestoreSearchResultByTag.from_response_json(
                result
            )
        elif filter_by == "tag_name":
            return search_results.FeaturestoreSearchResultByTagName.from_response_json(
                result
            )
        elif filter_by == "tag_key":
            return search_results.FeaturestoreSearchResultByTagKey.from_response_json(
                result
            )
        elif filter_by == "tag_value":
            return search_results.FeaturestoreSearchResultByTagValue.from_response_json(
                result
            )
        elif filter_by == "keyword":
            return search_results.FeaturestoreSearchResultByKeyWord.from_response_json(
                result
            )

        # Default case
        return search_results.FeaturestoreSearchResult.from_response_json(result)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

from enum import Enum
from typing import Any, Union

import requests
from requests.exceptions import SSLError


class RestAPIError(Exception):
    """REST Exception encapsulating the response object and url."""

    STATUS_CODE_BAD_REQUEST = 400
    STATUS_CODE_UNAUTHORIZED = 401
    STATUS_CODE_FORBIDDEN = 403
    STATUS_CODE_NOT_FOUND = 404
    STATUS_CODE_INTERNAL_SERVER_ERROR = 500

    class FeatureStoreErrorCode(int, Enum):
        FEATURE_GROUP_COMMIT_NOT_FOUND = 270227
        STATISTICS_NOT_FOUND = 270228

        def __eq__(self, other: Union[int, Any]) -> bool:
            if isinstance(other, int):
                return self.value == other
            if isinstance(other, self.__class__):
                return self is other
            return False

    def __init__(self, url: str, response: requests.Response) -> None:
        try:
            error_object = response.json()
            if isinstance(error_object, str):
                error_object = {"errorMsg": error_object}
        except Exception:
            error_object = {}
            self.error_code = None
        message = (
            "Metadata operation error: (url: {}). Server response: \n"
            "HTTP code: {}, HTTP reason: {}, body: {}, error code: {}, error msg: {}, user "
            "msg: {}".format(
                url,
                response.status_code,
                response.reason,
                response.content,
                error_object.get("errorCode", ""),
                error_object.get("errorMsg", ""),
                error_object.get("usrMsg", ""),
            )
        )
        if len(error_object) != 0:
            self.error_code = error_object.get("errorCode", "")
        super().__init__(message)
        self.url = url
        self.response = response


class UnknownSecretStorageError(Exception):
    """This exception will be raised if an unused secrets storage is passed as a parameter."""


class FeatureStoreException(Exception):
    """Generic feature store exception"""

    DUPLICATE_RECORD_ERROR_MESSAGE = (
        "Duplicate records detected: The dataset contains multiple rows that share identical values "
        "across all available columns from primary_key, and if defined: event_time and partition_key. "
        "Please remove or deduplicate these records before inserting."
    )


class VectorDatabaseException(Exception):
    # reason
    REQUESTED_K_TOO_LARGE = "REQUESTED_K_TOO_LARGE"
    REQUESTED_NUM_RESULT_TOO_LARGE = "REQUESTED_NUM_RESULT_TOO_LARGE"
    OTHERS = "OTHERS"

    # info
    REQUESTED_K_TOO_LARGE_INFO_K = "k"
    REQUESTED_NUM_RESULT_TOO_LARGE_INFO_N = "n"

    def __init__(self, reason: str, message: str, info: str) -> None:
        super().__init__(message)
        self._info = info
        self._reason = reason

    @property
    def reason(self) -> str:
        return self._reason

    @property
    def info(self) -> str:
        return self._info


class DataValidationException(FeatureStoreException):
    """Raised when data validation fails only when using "STRICT" validation ingestion policy."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class ExternalClientError(TypeError):
    """Raised when external client cannot be initialized due to missing arguments."""

    def __init__(self, missing_argument: str) -> None:
        message = (
            "{0} cannot be of type NoneType, {0} is a non-optional "
            "argument to connect to hopsworks from an external environment."
        ).format(missing_argument)
        super().__init__(message)


class InternalClientError(TypeError):
    """Raised when hopsworks internal client is missing some necessary configuration."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class HopsworksSSLClientError(SSLError):
    """Raised when the client connection fails with SSL related errors."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class GitException(Exception):
    """Generic git exception"""


class JobException(Exception):
    """Generic job exception"""


class EnvironmentException(Exception):
    """Generic python environment exception"""


class KafkaException(Exception):
    """Generic kafka exception"""


class DatasetException(Exception):
    """Generic dataset exception"""


class ProjectException(Exception):
    """Generic project exception"""


class OpenSearchException(Exception):
    """Generic opensearch exception"""


class JobExecutionException(Exception):
    """Generic job executions exception"""


class ModelRegistryException(Exception):
    """Generic model registry exception"""


class ModelServingException(Exception):
    """Generic model serving exception"""

    ERROR_CODE_SERVING_NOT_FOUND = 240000
    ERROR_CODE_ILLEGAL_ARGUMENT = 240001
    ERROR_CODE_DUPLICATED_ENTRY = 240011

    ERROR_CODE_DEPLOYMENT_NOT_RUNNING = 250001



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import warnings

from hopsworks_common import version
from hsfs import feature, feature_group, feature_view, training_dataset_feature
from hsfs.constructor import query
from hsfs.feature_store import FeatureStore
from hsfs.hopsworks_udf import udf
from hsfs.serving_key import ServingKey
from hsfs.transformation_function import TransformationType


fg1 = feature_group.FeatureGroup(
    name="test1",
    version=1,
    featurestore_id=99,
    primary_key=["primary_key"],
    event_time="event_time",
    partition_key=[],
    features=[
        feature.Feature("primary_key", primary=True, type="bigint"),
        feature.Feature("event_time", type="timestamp"),
        feature.Feature("fg1_feature", type="float"),
        feature.Feature("label", type="string"),
        feature.Feature("fg1_inference_helper", type="int"),
        feature.Feature("fg1_training_helper", type="int"),
    ],
    id=11,
    stream=False,
    featurestore_name="test_fs",
)

fg2 = feature_group.FeatureGroup(
    name="test2",
    version=1,
    featurestore_id=99,
    primary_key=["primary_key"],
    event_time="event_time",
    partition_key=[],
    features=[
        feature.Feature("primary_key", primary=True, type="bigint"),
        feature.Feature("event_time", type="timestamp"),
        feature.Feature("fg2_feature", type="float"),
        feature.Feature("fg2_inference_helper", type="int"),
        feature.Feature("fg2_training_helper", type="int"),
    ],
    id=11,
    stream=False,
    featurestore_name="test_fs",
)


class TestFeatureView:
    def test_from_response_json(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch.object(
            FeatureStore,
            "project_id",
            return_value=99,
        )
        mocker.patch("hopsworks_common.client.get_instance")
        mocker.patch("hsfs.engine.get_type")
        mocker.patch("hsfs.core.feature_store_api.FeatureStoreApi.get")
        json = backend_fixtures["feature_view"]["get"]["response"]
        # Act
        fv = feature_view.FeatureView.from_response_json(json)

        # Assert
        assert fv.name == "test_name"
        assert fv.id == 11
        assert isinstance(fv.query, query.Query)
        assert fv.featurestore_id == 5
        assert fv.version == 1
        assert fv.description == "test_description"
        assert fv.labels == ["intt"]
        assert fv.transformation_functions == []
        assert len(fv.schema) == 2
        assert isinstance(fv.schema[0], training_dataset_feature.TrainingDatasetFeature)

    def test_from_response_json_basic_info(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hsfs.engine.get_type")
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_view"]["get_basic_info"]["response"]

        # Act
        fv = feature_view.FeatureView.from_response_json(json)

        # Assert
        assert fv.name == "test_name"
        assert fv.id is None
        assert isinstance(fv.query, query.Query)
        assert fv.featurestore_id == 5
        assert fv.version is None
        assert fv.description is None
        assert fv.labels == []
        assert fv.transformation_functions == []
        assert len(fv.schema) == 0
        assert fv.query._left_feature_group.deprecated is False

    def test_from_response_json_transformation_function(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch.object(
            FeatureStore,
            "project_id",
            return_value=99,
        )
        mocker.patch("hopsworks_common.client.get_instance")
        mocker.patch("hsfs.engine.get_type")
        mocker.patch("hsfs.core.feature_store_api.FeatureStoreApi.get")
        json = backend_fixtures["feature_view"]["get_transformations"]["response"]
        # Act
        fv = feature_view.FeatureView.from_response_json(json)

        # Assert
        assert fv.name == "test_name"
        assert fv.id == 11
        assert isinstance(fv.query, query.Query)
        assert fv.featurestore_id == 5
        assert fv.version == 1
        assert fv.description == "test_description"
        assert fv.labels == ["intt"]
        assert len(fv.transformation_functions) == 2
        assert (
            fv.transformation_functions[0].hopsworks_udf.function_name == "add_mean_fs"
        )
        assert (
            fv.transformation_functions[1].hopsworks_udf.function_name == "add_one_fs"
        )
        assert (
            fv.transformation_functions[0].hopsworks_udf._function_source
            == "\n@udf(float)\ndef add_mean_fs(data1 : pd.Series, statistics=stats):\n    return data1 + statistics.data1.mean\n"
        )
        assert (
            fv.transformation_functions[1].hopsworks_udf._function_source
            == "\n@udf(float)\ndef add_one_fs(data1 : pd.Series):\n    return data1 + 1\n"
        )
        assert (
            fv.transformation_functions[0].transformation_type
            == TransformationType.MODEL_DEPENDENT
        )
        assert (
            fv.transformation_functions[1].transformation_type
            == TransformationType.MODEL_DEPENDENT
        )

        assert len(fv.schema) == 2
        assert isinstance(fv.schema[0], training_dataset_feature.TrainingDatasetFeature)

    def test_from_response_json_basic_info_deprecated(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hsfs.engine.get_type")
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_view"]["get_basic_info_deprecated"]["response"]

        # Act
        with warnings.catch_warnings(record=True) as warning_record:
            fv = feature_view.FeatureView.from_response_json(json)

        # Assert
        assert fv.name == "test_name"
        assert fv.id is None
        assert isinstance(fv.query, query.Query)
        assert fv.featurestore_id == 5
        assert fv.version is None
        assert fv.description is None
        assert fv.labels == []
        assert fv.transformation_functions == []
        assert len(fv.schema) == 0
        assert fv.query._left_feature_group.deprecated is True
        assert len(warning_record) == 1
        assert str(warning_record[0].message) == (
            f"Feature Group `{fv.query._left_feature_group.name}`, version `{fv.query._left_feature_group.version}` is deprecated"
        )

    def test_transformation_function_instances(self, mocker, backend_fixtures):
        # Arrange
        feature_store_id = 99
        mocked_connection = mocker.MagicMock()
        mocked_connection.backend_version = version.__version__
        mocked_connection = mocker.patch(
            "hopsworks_common.client.get_connection", return_value=mocked_connection
        )
        mocker.patch("hsfs.core.feature_view_engine.FeatureViewEngine")
        mocker.patch("hsfs.engine.get_type", return_value="python")
        json = backend_fixtures["query"]["get"]["response"]

        # Act
        q = query.Query.from_response_json(json)

        @udf(int)
        def test(col1):
            return col1 + 1

        fv = feature_view.FeatureView(
            featurestore_id=feature_store_id,
            name="test_fv",
            version=1,
            query=q,
            transformation_functions=[test("data1"), test("data2")],
        )

        transformation_functions = fv.transformation_functions

        assert transformation_functions[0] != transformation_functions[1]

    def test_logging_enabled_instances(self, mocker, backend_fixtures):
        # Arrange
        mocked_connection = mocker.MagicMock()
        mocked_connection.backend_version = version.__version__
        mocked_connection = mocker.patch(
            "hopsworks_common.client.get_connection", return_value=mocked_connection
        )
        mocker.patch("hsfs.core.feature_view_engine.FeatureViewEngine")
        mocker.patch("hsfs.engine.get_type", return_value="python")
        json = backend_fixtures["feature_view"]["get_feature_view_logging_enabled"][
            "response"
        ]

        # Act
        fv = feature_view.FeatureView.from_response_json(json)

        # Assert
        assert fv.logging_enabled is True

    def test_label_column_name(self, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        mocker.patch("hsfs.engine.get_type", return_value="python")

        fv = feature_view.FeatureView(
            name="fv_name",
            query=fg1.select_features().join(fg2.select_features()),
            featurestore_id=99,
            featurestore_name="test_fs",
            inference_helper_columns=["fg1_inference_helpers", "fg2_inference_helpers"],
            training_helper_columns=["fg1_training_helper", "fg2_training_helper"],
            labels=["label"],
        )

        # Act
        mocker.patch.object(
            fv,
            "get_training_dataset_schema",
            return_value=[
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_feature", type="float"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_inference_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_training_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="label", type="int", label=True
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_feature", type="float"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_inference_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_training_helper", type="int"
                ),
            ],
        )

        # Assert
        assert fv._label_column_names == {"label"}

    def test_transformed_feature_name(self, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        mocker.patch("hsfs.engine.get_type", return_value="python")

        fv = feature_view.FeatureView(
            name="fv_name",
            query=fg1.select_features().join(fg2.select_features()),
            featurestore_id=99,
            featurestore_name="test_fs",
            inference_helper_columns=["fg1_inference_helper", "fg2_inference_helper"],
            training_helper_columns=["fg1_training_helper", "fg2_training_helper"],
            labels=["label"],
        )

        mocker.patch.object(
            fv,
            "get_training_dataset_schema",
            return_value=[
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_feature", type="float"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_inference_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_training_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_transformed_features", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="label", type="int", label=True
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_feature", type="float"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_inference_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_training_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_transformed_features", type="int"
                ),
            ],
        )

        # Act
        transformed_features = fv._transformed_feature_names

        # Assert
        assert transformed_features == [
            "fg1_feature",
            "fg1_transformed_features",
            "fg2_feature",
            "fg2_transformed_features",
        ]

    def test_untransformed_feature_names(self, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        mocker.patch("hsfs.engine.get_type", return_value="python")

        fv = feature_view.FeatureView(
            name="fv_name",
            query=fg1.select_features().join(fg2.select_features()),
            featurestore_id=99,
            featurestore_name="test_fs",
            inference_helper_columns=["fg1_inference_helper", "fg2_inference_helper"],
            training_helper_columns=["fg1_training_helper", "fg2_training_helper"],
            labels=["label"],
        )

        mocker.patch.object(
            fv,
            "get_training_dataset_schema",
            return_value=[
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_feature", type="float"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_inference_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg1_training_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="label", type="int", label=True
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_feature", type="float"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_inference_helper", type="int"
                ),
                training_dataset_feature.TrainingDatasetFeature(
                    name="fg2_training_helper", type="int"
                ),
            ],
        )

        # Act
        transformed_features = fv._transformed_feature_names

        # Assert
        assert transformed_features == ["fg1_feature", "fg2_feature"]

    def test_required_serving_key_names(self, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        mocker.patch("hsfs.engine.get_type", return_value="python")

        fv = feature_view.FeatureView(
            name="fv_name",
            query=fg1.select_features().join(fg2.select_features()),
            featurestore_id=99,
            featurestore_name="test_fs",
            inference_helper_columns=["fg1_inference_helper", "fg2_inference_helper"],
            training_helper_columns=["fg1_training_helper", "fg2_training_helper"],
            labels=["label"],
        )

        fv._serving_keys = [
            ServingKey(
                feature_name="primary_key",
                join_index=0,
                feature_group=fg1,
                required=True,
            ),
            ServingKey(
                feature_name="primary_key",
                join_index=0,
                feature_group=fg2,
                required=False,
            ),
        ]

        # Act
        required_serving_keys = fv._required_serving_key_names

        # Assert
        assert required_serving_keys == ["primary_key"]

    def test_root_feature_group_event_time_column_name(self, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        mocker.patch("hsfs.engine.get_type", return_value="python")

        fv = feature_view.FeatureView(
            name="fv_name",
            query=fg1.select_features().join(fg2.select_features()),
            featurestore_id=99,
            featurestore_name="test_fs",
            inference_helper_columns=["fg1_inference_helper", "fg2_inference_helper"],
            training_helper_columns=["fg1_training_helper", "fg2_training_helper"],
            labels=["label"],
        )

        # Act
        root_feature_group_event_time = fv._root_feature_group_event_time_column_name

        # Assert
        assert root_feature_group_event_time == "event_time"



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from datetime import date, datetime
from typing import Optional, Union

from pydantic import BaseModel


class FeatureGroup(BaseModel):
    """Model representing a feature group in Hopsworks MCP."""

    id: Optional[int] = None
    name: Optional[str]
    version: Optional[int]
    featurestore_id: Optional[int]
    location: Optional[str]
    event_time: Optional[Union[str, int, date, datetime]] = None
    online_enabled: bool = False
    online_topic_name: Optional[str] = None
    topic_name: Optional[str] = None
    notification_topic_name: Optional[str] = None
    deprecated: bool = False


class FeatureGroups(BaseModel):
    """Model representing a collection of feature groups in Hopsworks MCP."""

    feature_groups: list[FeatureGroup]
    total: int
    offset: int
    limit: int



class TAGS:
    """
    This class contains constants for different tag types used in the system.
    """

    AUTH = "auth"
    PROJECT = "project"
    FEATURE_STORE = "feature_store"
    FEATURE_GROUP = "feature_group"
    FEATURE_VIEW = "feature_view"
    DATASET = "dataset"
    ENVIRONMENT = "environment"
    JOB = "job"
    SYSTEM = "system"

    STATEFUL = "stateful"
    STATELESS = "stateless"

    READ = "read"
    WRITE = "write"



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import hopsworks
from fastmcp import Context
from hopsworks.mcp.models.project import Project, Projects
from hopsworks.mcp.utils.tags import TAGS
from hopsworks_common import client


class ProjectResources:
    """Resources for Hopsworks projects."""

    def __init__(self, mcp):
        """Initialize project resources.

        Args:
            mcp: The MCP server instance
        """
        self.mcp = mcp
        self.mcp.resource(
            uri="project://projects",
            mime_type="application/json",
            tags=[TAGS.PROJECT, TAGS.STATELESS],
        )(self.list_projects)
        self.mcp.resource(
            uri="project://projects/{name}",
            mime_type="application/json",
            tags=[TAGS.PROJECT, TAGS.STATELESS],
        )(self.get_project_details)
        self.mcp.resource(
            uri="project://projects/current",
            mime_type="application/json",
            tags=[TAGS.PROJECT, TAGS.STATEFUL],
        )(self.get_current_project_details)

    async def list_projects(self, ctx: Context = None) -> Projects:
        """List all projects.

        Returns:
            Projects: A list of projects accessible by the user or an error message.
        """
        if ctx:
            await ctx.info("Listing all projects...")

        conn = client.get_connection()
        projects = conn.get_projects()
        return Projects(
            projects=[
                Project(
                    name=project.name,
                    id=project.id,
                    owner=project.owner,
                    description=project.description,
                    created=project.created,
                )
                for project in projects
            ]
        )

    async def get_current_project_details(self, ctx: Context = None) -> Project:
        """Get details of the current project.

        Returns:
            Project: The current project details or an error message.
        """
        if ctx:
            await ctx.info("Retrieving current project details...")

        project = hopsworks.get_current_project()
        return Project(
            name=project.name,
            id=project.id,
            owner=project.owner,
            description=project.description,
            created=project.created,
        )

    async def get_project_details(self, name: str, ctx: Context = None) -> Project:
        """Get project details.

        Args:
            name (str): The name of the project.

        Returns:
            Project: The project details or an error message.
        """
        if ctx:
            await ctx.info(f"Retrieving project details for {name}...")

        conn = client.get_connection()
        project = conn.get_project(name)
        return Project(
            name=project.name,
            id=project.id,
            owner=project.owner,
            description=project.description,
            created=project.created,
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import builtins
from typing import Literal, Optional, Union

import hopsworks


def login(
    host: Optional[str] = None,
    port: int = 443,
    project: Optional[str] = None,
    api_key_value: Optional[str] = None,
    api_key_file: Optional[str] = None,
    hostname_verification: bool = False,
    trust_store_path: Optional[str] = None,
    engine: Union[
        None,
        Literal["spark"],
        Literal["python"],
        Literal["training"],
        Literal["spark-no-metastore"],
        Literal["spark-delta"],
    ] = "python",
):
    """Login to Hopsworks and return the project.
    This function wraps hopsworks.login to simulate user input by overriding the built-in input function.

    Args:
        host (Optional[str]): Hopsworks host URL.
        port (int): Hopsworks port (default 443).
        project (Optional[str]): Project name to access.
        api_key_value (Optional[str]): API key value for Hopsworks authentication.
        api_key_file (Optional[str]): Path to a file containing the API key for Hopsworks authentication.
        hostname_verification (bool): Enable hostname verification for Hopsworks authentication.
        trust_store_path (Optional[str]): Path to the trust store for Hopsworks authentication.
        engine (Union[None, Literal["spark"], Literal["python"], Literal["training"], Literal["spark-no-metastore"], Literal["spark-delta"]]): Engine to use (default: python).
    Returns:
        Project: The project object after successful login.
    """
    # Backup original input
    original_input = builtins.input

    # Override input to simulate user input. Choose the first project by default.
    builtins.input = lambda _: 1

    try:
        project = hopsworks.login(
            host=host,
            port=port,
            api_key_value=api_key_value,
            api_key_file=api_key_file,
            hostname_verification=hostname_verification,
            trust_store_path=trust_store_path,
            engine=engine,
        )
    finally:
        # Restore original input function
        builtins.input = original_input

    return project



import datetime
import json
import warnings
from typing import Any, Dict, List, Optional

import humps
from hsfs import feature_group, util
from hsfs.feature import Feature


class LoggingMetaData:
    """
    Class that holds the data for feature logging.
    """

    def __init__(self):
        self.untransformed_features: Optional[list[list[Any]]] = []
        self.transformed_features: list[list[Any]] = []
        self.serving_keys: list[Dict[str, Any]] = []
        self.request_parameters: list[Dict[str, Any]] = []
        self.event_time: list[datetime.datetime] = []
        self.inference_helper: list[Dict[str, Any]] = []

    def __repr__(self):
        return (
            f"LoggingMetaData(untransformed_features={self.untransformed_features}, \n"
            f"transformed_features={self.transformed_features}, \n"
            f"serving_keys={self.serving_keys}, \n"
            f"request_parameters={self.request_parameters}, \n"
            f"event_time={self.event_time}, \n"
            f"inference_helper={self.inference_helper})"
        )


class FeatureLogging:
    NOT_FOUND_ERROR_CODE = 270248

    def __init__(
        self,
        id: int = None,
        transformed_features: "feature_group.FeatureGroup" = None,
        untransformed_features: "feature_group.FeatureGroup" = None,
        extra_logging_columns: Optional[List[Feature]] = None,
    ):
        """
        DTO class for feature logging.

        # Arguments
            id : `int`. Id of the feature logging object.
            transformed_features : `FeatureGroup`. The feature group containing the transformed features. As of Hopsworks 4.6, transformed and untransformed features are logged in the same feature group. This feature group is maintained for backward compatibility.
            untransformed_features : `FeatureGroup`. The feature group containing the untransformed features.
            extra_logging_columns : `List[Feature]`. List of extra logging columns.
        """
        self._id = id
        self._transformed_features = transformed_features
        self._untransformed_features = untransformed_features
        self._extra_logging_columns = extra_logging_columns

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> "FeatureLogging":
        from hsfs.feature_group import FeatureGroup  # avoid circular import

        json_decamelized = humps.decamelize(json_dict)
        transformed_features = json_decamelized.get("transformed_log_fg")
        untransformed_features = json_decamelized.get("untransformed_log_fg")
        if transformed_features:
            transformed_features = FeatureGroup.from_response_json(transformed_features)
        if untransformed_features:
            untransformed_features = FeatureGroup.from_response_json(
                untransformed_features
            )
        extra_logging_columns = json_decamelized.get("extra_logging_columns")
        if extra_logging_columns:
            extra_logging_columns = [
                Feature.from_response_json(feature) for feature in extra_logging_columns
            ]
        return cls(
            json_decamelized.get("id"),
            transformed_features,
            untransformed_features,
            extra_logging_columns,
        )

    def update(self, others):
        self._transformed_features = others.transformed_features
        self._untransformed_features = others.untransformed_features
        return self

    @property
    def transformed_features(self) -> "feature_group.FeatureGroup":
        return self._transformed_features

    @property
    def untransformed_features(self) -> "feature_group.FeatureGroup":
        return self._untransformed_features

    @property
    def extra_logging_columns(self) -> Optional[List[Feature]]:
        return self._extra_logging_columns

    def get_feature_group(self, transformed: Optional[bool] = None):
        if transformed is not None:
            warnings.warn(
                "Providing ´transformed´ while fetching logging feature group is deprecated"
                + " and will be dropped in future versions. Transformed and untransformed features are now logged in the same feature group.",
                DeprecationWarning,
                stacklevel=2,
            )
        if transformed:
            if self._transformed_features is None:
                return self._untransformed_features
            return self._transformed_features
        else:
            return self._untransformed_features

    @property
    def id(self) -> str:
        return self._id

    def to_dict(self):
        return {
            "id": self._id,
            "transformedLogFg": self._transformed_features,
            "untransformedLogFg": self._untransformed_features,
            "extraLoggingColumns": self._extra_logging_columns,
        }

    def json(self) -> Dict[str, Any]:
        return json.dumps(self, cls=util.Encoder)

    def __repr__(self):
        return self.json()



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from hopsworks_common import client
from hsfs.constructor import fs_query


class QueryConstructorApi:
    def construct_query(self, query):
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "featurestores", "query"]
        headers = {"content-type": "application/json"}
        return fs_query.FsQuery.from_response_json(
            _client._send_request(
                "PUT", path_params, headers=headers, data=query.json()
            )
        )



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os
from typing import Optional

import humps
from hopsworks_common import client, command, usage, util
from hopsworks_common.core import environment_api, library_api
from hopsworks_common.engine import environment_engine


class Environment:
    NOT_FOUND_ERROR_CODE = 300000

    def __init__(
        self,
        name=None,
        description=None,
        python_version=None,
        python_conflicts=None,
        pip_search_enabled=None,
        conflicts=None,
        conda_channel=None,
        libraries=None,
        commands=None,
        href=None,
        type=None,
        **kwargs,
    ):
        self._name = name
        self._description = description
        self._python_version = python_version
        self._python_conflicts = python_conflicts
        self._pip_search_enabled = pip_search_enabled
        self._conflicts = conflicts
        self._libraries = libraries
        self._commands = (
            command.Command.from_response_json(commands) if commands else None
        )

        self._environment_engine = environment_engine.EnvironmentEngine()
        self._library_api = library_api.LibraryApi()
        self._environment_api = environment_api.EnvironmentApi()

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            return [cls(**env) for env in json_decamelized["items"]]
        else:
            return cls(**json_decamelized)

    @property
    def python_version(self):
        """Python version of the environment"""
        return self._python_version

    @property
    def name(self):
        """Name of the environment"""
        return self._name

    @property
    def description(self):
        """Description of the environment"""
        return self._description

    @usage.method_logger
    def install_wheel(self, path: str, await_installation: Optional[bool] = True):
        """Install a python library packaged in a wheel file

        ```python

        import hopsworks

        project = hopsworks.login()

        # Upload to Hopsworks
        ds_api = project.get_dataset_api()
        whl_path = ds_api.upload("matplotlib-3.1.3-cp38-cp38-manylinux1_x86_64.whl", "Resources")

        # Install
        env_api = project.get_environment_api()
        env = env_api.get_environment("my_custom_environment")

        env.install_wheel(whl_path)

        ```

        # Arguments
            path: str. The path on Hopsworks where the wheel file is located
            await_installation: bool. If True the method returns only when the installation finishes. Default True
        # Returns
            `Library`: The library object
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        # Wait for any ongoing environment operations
        self._environment_engine.await_environment_command(self.name)

        library_name = os.path.basename(path)

        _client = client.get_instance()
        path = util.convert_to_abs(path, _client._project_name)

        library_spec = {
            "dependencyUrl": path,
            "channelUrl": "wheel",
            "packageSource": "WHEEL",
        }

        library_rest = self._library_api._install(library_name, self.name, library_spec)

        if await_installation:
            return self._environment_engine.await_library_command(
                self.name, library_name
            )

        return library_rest

    @usage.method_logger
    def install_requirements(
        self, path: str, await_installation: Optional[bool] = True
    ):
        """Install libraries specified in a requirements.txt file

        ```python

        import hopsworks

        project = hopsworks.login()

        # Upload to Hopsworks
        ds_api = project.get_dataset_api()
        requirements_path = ds_api.upload("requirements.txt", "Resources")

        # Install
        env_api = project.get_environment_api()
        env = env_api.get_environment("my_custom_environment")


        env.install_requirements(requirements_path)

        ```

        # Arguments
            path: str. The path on Hopsworks where the requirements.txt file is located
            await_installation: bool. If True the method returns only when the installation is finished. Default True
        # Returns
            `Library`: The library object
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        # Wait for any ongoing environment operations
        self._environment_engine.await_environment_command(self.name)

        library_name = os.path.basename(path)

        _client = client.get_instance()
        path = util.convert_to_abs(path, _client._project_name)

        library_spec = {
            "dependencyUrl": path,
            "channelUrl": "requirements_txt",
            "packageSource": "REQUIREMENTS_TXT",
        }

        library_rest = self._library_api._install(library_name, self.name, library_spec)

        if await_installation:
            return self._environment_engine.await_library_command(
                self.name, library_name
            )

        return library_rest

    @usage.method_logger
    def delete(self):
        """Delete the environment
        !!! danger "Potentially dangerous operation"
            This operation deletes the python environment.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        self._environment_api._delete(self.name)

    def __repr__(self):
        return f"Environment({self.name!r})"



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import asyncio

import hopsworks
from fastmcp import Context
from hopsworks.mcp.models.project import Project, Projects
from hopsworks.mcp.utils.tags import TAGS
from hopsworks_common import client


class ProjectTools:
    """Tools for Hopsworks projects."""

    def __init__(self, mcp):
        """Initialize project tools.

        Args:
            mcp: The MCP server instance
        """
        self.mcp = mcp
        self.mcp.tool(tags=[TAGS.PROJECT, TAGS.READ, TAGS.STATEFUL])(self.use_project)
        self.mcp.tool(tags=[TAGS.PROJECT, TAGS.WRITE, TAGS.STATELESS])(
            self.create_project
        )
        self.mcp.tool(tags=[TAGS.PROJECT, TAGS.READ, TAGS.STATELESS])(
            self.list_projects
        )
        self.mcp.tool(tags=[TAGS.PROJECT, TAGS.READ, TAGS.STATEFUL])(
            self.get_current_project_details
        )
        self.mcp.tool(tags=[TAGS.PROJECT, TAGS.READ, TAGS.STATELESS])(
            self.get_project_details
        )

    async def _create_project(
        self, conn, name: str = None, description: str = None
    ) -> dict:
        loop = asyncio.get_event_loop()
        project = await loop.run_in_executor(
            None, conn.create_project, name, description
        )
        return project

    async def use_project(self, name: str, ctx: Context = None) -> Project:
        """Use a specific project.

        Args:
            name (str): The name of the project to use.
        Returns:
            Project: The project details or an error message.
        """
        if ctx:
            await ctx.info(f"Changing to project {name}...")

        conn = client.get_connection()
        if conn is None:
            raise ConnectionError("Not connected to Hopsworks.")

        project = hopsworks.login(
            host=conn.host,
            port=conn.port,
            project=name,
            api_key_value=conn.api_key_value,
            engine=conn._engine,
        )
        return Project(
            name=project.name,
            id=project.id,
            owner=project.owner,
            description=project.description,
            created=project.created,
        )

    async def create_project(
        self, name: str = None, description: str = None, ctx: Context = None
    ) -> Project:
        """Create a new project.

        Args:
            name (str): The name of the project.
            description (str): A description of the project.
        Returns:
            Project: The newly created project details or an error message.
        """
        if ctx:
            await ctx.info(f"Creating project {name}...")

        progress = 0
        conn = client.get_connection()
        task = asyncio.create_task(
            self._create_project(conn, name=name, description=description)
        )
        if ctx:
            while not task.done():
                await ctx.report_progress(
                    progress=progress, message=f"Creating project: {name}..."
                )
                progress += 1
                await asyncio.sleep(1)
            await ctx.report_progress(
                progress=progress, message=f"Project '{name}' created successfully."
            )

        project = await task
        return Project(
            name=project.name,
            id=project.id,
            owner=project.owner,
            description=project.description,
            created=project.created,
        )

    async def list_projects(self, ctx: Context = None) -> Projects:
        """List all projects.

        Returns:
            Projects: A list of projects accessible by the user or an error message.
        """
        if ctx:
            await ctx.info("Listing all projects...")

        conn = client.get_connection()
        projects = conn.get_projects()

        return Projects(
            projects=[
                Project(
                    name=project.name,
                    id=project.id,
                    owner=project.owner,
                    description=project.description,
                    created=project.created,
                )
                for project in projects
            ]
        )

    async def get_current_project_details(self, ctx: Context = None) -> Project:
        """Get details of the current project.

        Returns:
            Project: The current project details or an error message.
        """
        if ctx:
            await ctx.info("Retrieving current project details...")

        project = hopsworks.get_current_project()
        return Project(
            name=project.name,
            id=project.id,
            owner=project.owner,
            description=project.description,
            created=project.created,
        )

    async def get_project_details(self, name: str, ctx: Context = None) -> Project:
        """Get project details.

        Args:
            name (str): The name of the project.

        Returns:
            Project: The project details or an error message.
        """
        if ctx:
            await ctx.info(f"Retrieving project details for {name}...")

        conn = client.get_connection()
        project = conn.get_project(name)
        return Project(
            name=project.name,
            id=project.id,
            owner=project.owner,
            description=project.description,
            created=project.created,
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import logging
import signal
import sys
from typing import Literal, Optional, Union

import click
import hopsworks

from .server import mcp
from .utils.auth import login


# Configure logging to handle closed streams gracefully
class SafeStreamHandler(logging.StreamHandler):
    def emit(self, record):
        try:
            if self.stream and not getattr(self.stream, "closed", False):
                super().emit(record)
        except (ValueError, BrokenPipeError, OSError):
            print("Error writing log message to stream, stream may be closed.")
            # Ignore errors from closed streams
            pass


log = logging.getLogger("hopsworks-mcp")


def handle_shutdown(signum, frame):
    log.info(f"Received signal {signum}, will shut down...")
    hopsworks.logout()  # Checks if a session is active and logs out
    sys.exit(0)


@click.command(
    "hopsworks-mcp",
    short_help="Run the Hopsworks MCP server.",
    help="In addition to setting arguments directly, you can set environment variables: \
        HOPSWORKS_HOST, HOPSWORKS_PORT, HOPSWORKS_PROJECT, HOPSWORKS_API_KEY, \
            HOPSWORKS_HOSTNAME_VERIFICATION, HOPSWORKS_TRUST_STORE_PATH and HOPSWORKS_ENGINE.",
)
@click.option(
    "--host", default="0.0.0.0", help="Host to run the server on. (default: 0.0.0.0)"
)
@click.option("--port", default=8001, help="Port to run the server on. (default: 8001)")
@click.option(
    "--transport",
    default="sse",
    help="Transport method to use. (default: sse). Options: 'stdio', 'http', 'sse', 'streamable-http'",
)
@click.option(
    "--create_session",
    default=True,
    help="Create a Hopsworks session for the MCP server. (default: True)",
)
@click.option("--hopsworks_host", default=None, help="Hopsworks host URL")
@click.option("--hopsworks_port", default=443, help="Hopsworks port (default 443)")
@click.option("--project", default=None, help="Project name to access")
@click.option(
    "--api_key_value",
    default=None,
    help="API key value for Hopsworks authentication",
)
@click.option(
    "--api_key_file",
    default=None,
    help="Path to a file containing the API key for Hopsworks authentication",
)
@click.option(
    "--hostname_verification",
    default=False,
    help="Enable hostname verification for Hopsworks authentication",
)
@click.option(
    "--trust_store_path",
    default=None,
    help="Path to the trust store for Hopsworks authentication",
)
@click.option(
    "--engine",
    default="python",
    help="Engine to use (python, spark, training, spark-no-metastore, spark-delta) (default: python)",
)
def main(
    host: Optional[str] = None,
    port: int = 8001,
    transport: Optional[str] = None,
    create_session: bool = True,
    hopsworks_host: Optional[str] = None,
    hopsworks_port: int = 443,
    project: Optional[str] = None,
    api_key_value: Optional[str] = None,
    api_key_file: Optional[str] = None,
    hostname_verification: bool = False,
    trust_store_path: Optional[str] = None,
    engine: Union[
        None,
        Literal["spark"],
        Literal["python"],
        Literal["training"],
        Literal["spark-no-metastore"],
        Literal["spark-delta"],
    ] = "python",
):
    """Run the Hopsworks MCP server."""
    if transport not in ["stdio", "http", "sse", "streamable-http"]:
        raise ValueError(
            "Invalid transport type. Choose from 'stdio', 'http', 'sse', or 'streamable-http'."
        )

    if create_session:
        # Set the API key for the Hopsworks client
        project = login(
            host=hopsworks_host,
            port=hopsworks_port,
            project=project,
            api_key_value=api_key_value,
            api_key_file=api_key_file,
            hostname_verification=hostname_verification,
            trust_store_path=trust_store_path,
            engine=engine,
        )

    signal.signal(signal.SIGTERM, handle_shutdown)
    signal.signal(signal.SIGINT, handle_shutdown)

    if transport == "stdio":
        # For stdio transport, suppress all logging from imported libraries
        # that might try to write to closed streams
        logging.getLogger().handlers.clear()
        logging.getLogger().addHandler(SafeStreamHandler(sys.stdout))

        log.info(f"Starting Hopsworks MCP server using {transport} transport.")
        mcp.run(transport=transport, show_banner=False)
    else:
        mcp.run(transport=transport, host=host, port=port, show_banner=False)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""MCP server for Hopsworks."""

from fastmcp import FastMCP

from .prompts import ProjectPrompts, SystemPrompts
from .resources import FeatureStoreResources, ProjectResources
from .tools import (
    AuthTools,
    DatasetTools,
    FeatureGroupTools,
    FeatureStoreTools,
    JobTools,
    ProjectTools,
)


# Create a FastMCP server instance
mcp = FastMCP(name="Hopsworks MCP")

# Initialize tools and resources
_auth_tools = AuthTools(mcp)
_project_tools = ProjectTools(mcp)
_feature_store_tools = FeatureStoreTools(mcp)
_feature_store_resources = FeatureStoreResources(mcp)
_project_resource = ProjectResources(mcp)
_project_prompts = ProjectPrompts(mcp)
_system_prompts = SystemPrompts(mcp)
_job_tools = JobTools(mcp)
_dataset_tools = DatasetTools(mcp)
_feature_group_tools = FeatureGroupTools(mcp)
